"""Definitions of key functions"""

from __future__ import annotations

import logging as _logging
import os
from os.path import abspath, basename
from time import sleep
from types import FunctionType
from typing import Tuple

import urwid

from .. import __version__, logging
from ..config import context_keys, expand_key
from ..utils import get_terminal_size
from . import main
from .widgets import (
    ImageCanvas,
    bottom_bar,
    confirmation,
    confirmation_overlay,
    expand,
    image_box,
    image_grid,
    image_grid_box,
    key_bar,
    main as main_widget,
    menu,
    menu_box,
    overlay,
    pile,
    placeholder,
    view,
    viewer,
)

# Action Status Modification


def disable_actions(context: str, *actions: str) -> None:
    keyset = context_keys[context]
    for action in actions:
        keyset[action][4] = False
        keys[context][keyset[action][0]][1] = False
        display_context_keys(context)


def enable_actions(context: str, *actions: str) -> None:
    keyset = context_keys[context]
    for action in actions:
        keyset[action][4] = True
        keys[context][keyset[action][0]][1] = True
        display_context_keys(context)


def hide_actions(context: str, *actions: str) -> None:
    keyset = context_keys[context]
    for action in actions:
        keyset[action][3] = False
        display_context_keys(context)
    disable_actions(context, *actions)


def show_actions(context: str, *actions: str) -> None:
    keyset = context_keys[context]
    for action in actions:
        keyset[action][3] = True
        display_context_keys(context)
    enable_actions(context, *actions)


# Main


def change_key(context: str, old: str, new: str) -> None:
    """Changes the key for a registered action from *old* to *new*.

    Raises:
        KeyError: *old* was not registered.
    """
    keys[context][new] = keys[context].pop(old)


def display_context_help(context: str) -> None:
    """Displays the help menu for a particular context, showing all visible actions
    and their descriptions.
    """
    global _prev_view_widget

    actions = (
        *context_keys[context].items(),
        *(() if context in no_globals else context_keys["global"].items()),
    )

    separator = (1, urwid.Filler(urwid.Text("\u2502" * 3)))
    contents = [
        (
            3,
            urwid.Columns(
                [
                    (
                        "weight",
                        3,
                        urwid.Filler(
                            urwid.Text(("default bold", f"{action}"), "center")
                        ),
                    ),
                    separator,
                    (
                        "weight",
                        2,
                        urwid.Filler(
                            urwid.Text(("default bold", f"{symbol} ({key})"), "center")
                        ),
                    ),
                    separator,
                    (
                        "weight",
                        5,
                        urwid.Filler(
                            urwid.Text(("default bold", f"{description}"), "center")
                        ),
                    ),
                ],
                min_width=5,
            ),
        )
        for action, (key, symbol, description, visible, _) in actions
        if visible
    ]

    line = urwid.SolidFill("\u2500")
    divider = urwid.Columns(
        [
            ("weight", 3, line),
            (1, urwid.Filler(urwid.Text("\u253c"))),
            ("weight", 2, line),
            (1, urwid.Filler(urwid.Text("\u253c"))),
            ("weight", 5, line),
        ],
        min_width=5,
    )
    for index in range(1, (len(contents) - 1) * 2, 2):
        contents.insert(index, (1, divider))

    contents.insert(
        0,
        (
            1,
            urwid.Columns(
                [
                    ("weight", 3, line),
                    (1, urwid.Filler(urwid.Text("\u252c"))),
                    ("weight", 2, line),
                    (1, urwid.Filler(urwid.Text("\u252c"))),
                    ("weight", 5, line),
                ],
                min_width=5,
            ),
        ),
    )

    contents.extend(
        [
            (
                1,
                urwid.Columns(
                    [
                        ("weight", 3, line),
                        (1, urwid.Filler(urwid.Text("\u2534"))),
                        ("weight", 2, line),
                        (1, urwid.Filler(urwid.Text("\u2534"))),
                        ("weight", 5, line),
                    ],
                    min_width=5,
                ),
            ),
            (
                3,
                urwid.LineBox(
                    urwid.Filler(
                        urwid.Text(("default bold", f"Version {__version__}"), "center")
                    )
                ),
            ),
        ]
    )

    overlay.top_w.original_widget.body[0] = urwid.Pile(contents)
    overlay.bottom_w = view if main.get_context() == "full-image" else pile
    main_widget.contents[0] = (overlay, ("weight", 1))
    main.set_context("overlay")

    # `Image` widgets don't support overlay.
    # Always reset by or "overlay::Close"
    _prev_view_widget = view.original_widget
    view.original_widget = urwid.LineBox(
        placeholder, _prev_view_widget.title_widget.text.strip(" "), "left"
    )


def display_context_keys(context: str) -> None:
    """Updates the Key/Action bar with the actions in the given context.

    Includes "global" actions for all contexts except those in `no_globals`.
    """
    actions = (
        *context_keys[context].items(),
        *(() if context in no_globals else context_keys["global"].items()),
    )

    # The underscores and blocks (U+2588) are to prevent wrapping amidst keys
    key_bar.original_widget.set_text(
        [
            [
                ("key" if enabled else "disabled key", action.replace(" ", "\u2800")),
                ("key", "\u2800"),
                ("key" if enabled else "disabled key", f"[{symbol}]"),
                " ",
            ]
            for action, (_, symbol, _, visible, enabled) in actions
            if visible
        ]
    )
    resize()


def register_key(*args: Tuple[str, str]) -> FunctionType:
    """Returns a decorator to register a function to some context action(s).

    Args: `(context, action)` tuple(s), each specifying an *action* and it's *context*.

    Each *context* and *action* must be valid.
    If no argument is given, the wrapper simply does nothing.
    """

    def register(func: FunctionType) -> None:
        """Registers *func* to the key corresponding to each ``(context, action)`` pair
        recieved by the call to ``register_key()`` that defines it.
        """
        for context, action in args:
            # All actions are enabled by default
            keys[context][context_keys[context][action][0]] = [func, True]

        return func

    return register


def set_confirmation(
    msg: str,
    bottom_widget: urwid.widget,
    confirm: FunctionType,
    cancel: FunctionType,
    confirm_args: tuple = (),
    cancel_args: tuple = (),
) -> None:
    """Setup a confirmation dialog

    Args:
      - msg: The message to be displayed in the dialog.
      - bottom_widget: The widget on which the confirmation dialog will be overlayed.
      - confirm: A function to be called for the "Confirm" action of the
        confirmation context.
      - cancel: A function to be called for the "Cancel" action of the
        confirmation context.
      - confirm_args: Optional positional arguments to be passed to _confirm_.
      - cancel_args: Optional positional arguments to be passed to _cancel_.

    This function must be called by any context action using the confirmation dialog.
    """
    global _confirm, _cancel, _prev_view_widget

    _confirm = (confirm, confirm_args)
    _cancel = (cancel, cancel_args)
    confirmation.set_text(msg)
    main.set_context("confirmation")

    # `Image` widgets don't support overlay.
    # Always reset by or "confirmation::Cancel"
    # but _confirm()_ must reset `view.original_widget` on it's own.
    _prev_view_widget = view.original_widget
    view.original_widget = urwid.LineBox(
        placeholder, _prev_view_widget.title_widget.text.strip(" "), "left"
    )

    confirmation_overlay.bottom_w = bottom_widget
    main_widget.contents[0] = (confirmation_overlay, ("weight", 1))

    getattr(main.ImageClass, "clear", lambda: True)()


# Context Actions

# {<context>: {<key>: [<func>, <state>], ...}, ...}
keys = {context: {} for context in context_keys}


# global
@register_key(("global", "Quit"))
def quit():
    main.quitting.set()
    raise urwid.ExitMainLoop()


@register_key(("global", "Key Bar"))
def expand_collapse_keys():
    if expand._ti_shown:
        if key_bar._ti_collapsed and key_bar_rows() > 1:
            expand.original_widget.set_text(f"\u25BC [{expand_key[1]}]")
            main_widget.contents[-1] = (
                bottom_bar,
                ("given", key_bar_rows()),
            )
            key_bar._ti_collapsed = False
            getattr(main.ImageClass, "clear", lambda: True)() or ImageCanvas.change()
        elif not key_bar._ti_collapsed:
            expand.original_widget.set_text(f"\u25B2 [{expand_key[1]}]")
            main_widget.contents[-1] = (bottom_bar, ("given", 1))
            key_bar._ti_collapsed = True
            getattr(main.ImageClass, "clear", lambda: True)() or ImageCanvas.change()


@register_key(("global", "Help"))
def help():
    display_context_help(main.get_context())
    getattr(main.ImageClass, "clear", lambda: True)()


def resize():
    cols = get_terminal_size()[0]
    rows = key_bar.original_widget.rows((cols,))
    if expand._ti_shown:
        if rows == 1:
            bottom_bar.contents.pop()
            expand._ti_shown = False
    elif rows > 1:
        bottom_bar.contents.append(
            (expand, ("given", len(expand.original_widget.text), False))
        )
        expand._ti_shown = True

    if not key_bar._ti_collapsed:
        new_rows = key_bar_rows()
        if main_widget.contents[-1][1][1] != new_rows:
            getattr(main.ImageClass, "clear", lambda: True)()
        main_widget.contents[-1] = (
            bottom_bar,
            ("given", new_rows),
        )


def key_bar_rows():
    # Consider columns occupied by the expand key and the divider
    cols = (
        get_terminal_size()[0]
        - (len(expand.original_widget.text) + 2) * expand._ti_shown
    )
    return key_bar.original_widget.rows((cols,))


keys["global"].update({"resized": [resize, True]})


# menu
@register_key(
    ("menu", "Prev"),
    ("menu", "Next"),
    ("menu", "Page Up"),
    ("menu", "Page Down"),
    ("menu", "Top"),
    ("menu", "Bottom"),
)
def menu_nav():
    main.displayer.send(menu.focus_position - 1)
    if not main.at_top_level or main.menu_list:
        set_menu_actions()
        set_menu_count()


def set_menu_actions():
    pos = menu.focus_position - 1
    if pos == -1:
        disable_actions("menu", "Switch Pane", "Delete", "Prev", "Page Up", "Top")
    elif main.menu_list[pos][1] is ...:
        disable_actions("menu", "Delete")
        enable_actions("menu", "Prev", "Page Up", "Top")
    else:
        enable_actions("menu", "Switch Pane", "Delete", "Prev", "Page Up", "Top")

    if main.at_top_level:
        if pos == 0:
            # "Top" is not disabled to ensure ".." is never selected
            # See `pos == -1` in `.main.display_images()`
            disable_actions("menu", "Prev", "Page Up")
        disable_actions("menu", "Back")
    else:
        enable_actions("menu", "Back")

    if main.menu_scan_done.is_set() and pos == len(main.menu_list) - 1:
        disable_actions("menu", "Next", "Page Down", "Bottom")
    else:
        enable_actions("menu", "Next", "Page Down", "Bottom")


def set_menu_count():
    length = len(main.menu_list) if main.menu_scan_done.is_set() else "..."
    menu_box.set_title(f"{menu.focus_position} of {length}")


@register_key(("menu", "Open"))
def open():
    if menu.focus_position == 0 or main.menu_list[menu.focus_position - 1][1] is ...:
        main.displayer.send(main.OPEN)
    else:
        main.set_context("full-image")
        main_widget.contents[0] = (view, ("weight", 1))
        set_image_view_actions()

    getattr(main.ImageClass, "clear", lambda: True)()


@register_key(("menu", "Back"))
def back():
    main.displayer.send(main.BACK)
    getattr(main.ImageClass, "clear", lambda: True)()


# image
@register_key(("image", "Maximize"))
def maximize():
    main.set_context("full-image")
    main_widget.contents[0] = (view, ("weight", 1))
    set_image_view_actions()

    getattr(main.ImageClass, "clear", lambda: True)()


# image-grid
@register_key(("image-grid", "Size-"))
def cell_width_dec():
    if image_grid.cell_width > 30:
        image_grid.cell_width -= 2
        main.grid_render_queue.put(None)  # Mark the start of a new grid
        main.grid_change.set()
        # Wait till GridRenderManager clears the cache
        while main.grid_change.is_set():
            pass
        getattr(main.ImageClass, "clear", lambda: True)()


@register_key(("image-grid", "Size+"))
def cell_width_inc():
    if image_grid.cell_width < 50:
        image_grid.cell_width += 2
        main.grid_render_queue.put(None)  # Mark the start of a new grid
        main.grid_change.set()
        # Wait till GridRenderManager clears the cache
        while main.grid_change.is_set():
            pass
        getattr(main.ImageClass, "clear", lambda: True)()


@register_key(("image-grid", "Open"))
def maximize_cell():
    main.set_context("full-grid-image")
    row = image_grid_box.base_widget.focus
    image_w = (
        row.focus
        if isinstance(row, urwid.Columns)  # when maxcol >= cell_width
        else row
    ).original_widget.original_widget  # The Image is in a LineSquare in an AttrMap

    image_box._w.contents[1][0].contents[1] = (image_w, ("weight", 1, True))
    image_box.set_title(basename(image_w._ti_image._source))
    main_widget.contents[0] = (image_box, ("weight", 1))

    image_box.original_widget = image_w
    if image_w._ti_image._is_animated:
        main.animate_image(image_w)

    getattr(main.ImageClass, "clear", lambda: True)()


def set_image_grid_actions():
    # The grid for a non-empty directory might be empty at the start of scanning
    if image_grid.contents:
        enable_actions("image-grid", "Open", "Size-", "Size+")
    else:
        disable_actions("image-grid", "Open", "Size-", "Size+")


# full-grid-image
@register_key(("full-grid-image", "Force Render"))
def force_render_maximized_cell():
    # Will re-render immediately after processing input, since caching has been disabled
    # for `Image` widgets.
    image_w = image_box._w.contents[1][0].contents[1][0]
    if image_w._ti_image._is_animated:
        main.animate_image(image_w, True)
    else:
        image_w._ti_force_render = True


# full-image, full-grid-image
@register_key(("full-image", "Restore"), ("full-grid-image", "Back"))
def restore():
    if main.get_context() == "full-grid-image":
        image_box.original_widget = placeholder  # halt image and anim rendering

    main.set_prev_context()
    main_widget.contents[0] = (pile, ("weight", 1))
    if main.get_context() == "menu":
        set_menu_actions()
    elif main.get_context() == "image":
        set_image_view_actions()

    getattr(main.ImageClass, "clear", lambda: True)()


# image, full-image
@register_key(("image", "Prev"), ("full-image", "Prev"))
def prev_image():
    if (
        menu.focus_position > 1
        # Don't scroll through directory items in image views
        and main.menu_list[menu.focus_position - 2][1] is not ...  # Previous item
    ):
        menu.focus_position -= 1
        main.displayer.send(menu.focus_position - 1)

    set_image_view_actions()
    set_menu_count()


@register_key(("image", "Next"), ("full-image", "Next"))
def next_image():
    # `menu_list` is one item less than `menu` (at it's beginning), hence no `len - 1`
    if menu.focus_position < len(main.menu_list):
        menu.focus_position += 1
        main.displayer.send(menu.focus_position - 1)

    set_image_view_actions()
    set_menu_count()


@register_key(("image", "Force Render"), ("full-image", "Force Render"))
def force_render():
    # Will re-render immediately after processing input, since caching has been disabled
    # for `Image` widgets.
    image_w = main.menu_list[menu.focus_position - 1][1]
    if image_w._ti_image._is_animated:
        main.animate_image(image_w, True)
    else:
        image_w._ti_force_render = True


def set_image_view_actions(context: str = None):
    context = context or main.get_context()
    if (
        menu.focus_position < 2
        # Previous item is a directory
        or main.menu_list[menu.focus_position - 2][1] is ...
    ):
        disable_actions(context, "Prev")
    else:
        enable_actions(context, "Prev")

    if (
        # Last item
        main.menu_scan_done.is_set()
        and menu.focus_position == len(main.menu_list)
    ):
        disable_actions(context, "Next")
    else:
        enable_actions(context, "Next")


# menu, image, full-image
@register_key(
    ("menu", "Delete"),
    ("image", "Delete"),
    ("full-image", "Delete"),
)
def delete():
    entry = main.menu_list[menu.focus_position - 1][0]
    set_confirmation(
        ("warning", "Permanently delete this image?"),
        view if main.get_context() == "full-image" else pile,
        _confirm_delete,
        _cancel_delete,
        (entry,),
    )


def _confirm_delete(entry):
    try:
        os.remove(entry)
    except OSError:
        successful = False
        logging.log_exception(f"Unable to delete {abspath(entry)!r}", logger)
        confirmation.set_text(("warning", "Unable to delete! Check the logs for info."))
    else:
        successful = True
        main.displayer.send(main.DELETE)
        confirmation.set_text(f"Successfully deleted {abspath(entry)}")
        confirmation.set_text(("green fg", "Successfully deleted!"))
    main.loop.draw_screen()
    sleep(1)

    if successful:
        next(main.displayer)  # Render next image view
        if not main.menu_list or main.menu_list[menu.focus_position - 1][1] is ...:
            # All menu entries have been deleted OR selected menu item is a directory
            main_widget.contents[0] = (pile, ("weight", 1))
            viewer.focus_position = 0
            # "confirmation:Confirm" calls `set_prev_context()`
            main._prev_contexts[0] = (
                "global" if main.at_top_level and not main.menu_list else "menu"
            )
        else:
            _cancel_delete()
    else:
        view.original_widget = _prev_view_widget
        _cancel_delete()


def _cancel_delete():
    main_widget.contents[0] = (
        view if main.get_prev_context() == "full-image" else pile,
        ("weight", 1),
    )
    if main.get_prev_context() in {"image", "full-image"}:
        set_image_view_actions(main.get_prev_context())


# menu, image, image-grid
@register_key(
    ("menu", "Switch Pane"),
    ("image", "Switch Pane"),
    ("image-grid", "Switch Pane"),
)
def switch_pane():
    if main.get_context() != "menu":
        main.set_context("menu")
        viewer.focus_position = 0
        set_menu_actions()
    else:
        viewer.focus_position = 1
        if view.original_widget is image_box:
            main.set_context("image")
            set_image_view_actions()
        else:
            main.set_context("image-grid")
            set_image_grid_actions()


# confirmation
@register_key(("confirmation", "Confirm"))
def confirm():
    # `_confirm()` must [re]set `view.original_widget`
    _confirm[0](*_confirm[1])
    main.set_prev_context()


@register_key(("confirmation", "Cancel"))
def cancel():
    _cancel[0](*_cancel[1])
    view.original_widget = _prev_view_widget
    main.set_prev_context()


# overlay
@register_key(("overlay", "Close"))
def close():
    main_widget.contents[0] = (
        view if main.get_prev_context() == "full-image" else pile,
        ("weight", 1),
    )
    view.original_widget = _prev_view_widget
    main.set_prev_context()


logger = _logging.getLogger(__name__)
no_globals = {"global", "confirmation", "full-grid-image", "overlay"}
key_bar._ti_collapsed = True
expand._ti_shown = True

# The annotations below are put in comments for compatibility with Python 3.7
# as it doesn't allow names declared as `global` within functions to be annotated.

# Use in the "confirmation" context. Set by `set_confirmation()`
_confirm = None  #: Optional[Tuple[FunctionType, Tuple[Any]]]
_cancel = None  #: Optional[Tuple[FunctionType, Tuple[Any]]]

# Used for overlays
_prev_view_widget = None  #: Optional[urwid.Widget]



"""
term-image

Display images in the terminal

Provides
========
    1. A library with features to display images in a terminal in various ways.
    2. A CLI to display images from a local filesystem or URLs.
    3. A TUI to browse through images and directories on a local filesystem
       or from URLS.

AUTHOR: AnonymouX47 <anonymoux47@gmail.com>
Copyright (c) 2022
"""

from __future__ import annotations

__all__ = ("set_cell_ratio", "get_cell_ratio", "AutoCellRatio")
__author__ = "AnonymouX47"

from enum import Enum, auto
from operator import truediv
from typing import Optional, Union

from .exceptions import TermImageError
from .utils import get_cell_size

version_info = (0, 5, 0)
__version__ = ".".join(map(str, version_info))


def get_cell_ratio() -> float:
    """Returns the global :term:`cell ratio`.

    See :py:func:`set_cell_ratio`.
    """
    # `(1, 2)` is a fallback in case the terminal doesn't respond in time
    return _cell_ratio or truediv(*(get_cell_size() or (1, 2)))


def set_cell_ratio(ratio: Union[float, AutoCellRatio]) -> None:
    """Sets the global :term:`cell ratio`.

    Args:
        ratio: Can be one of the following values.

          * A positive ``float`` value.
          * :py:attr:`AutoCellRatio.FIXED`, the ratio is immediately determined from
            the :term:`active terminal`.
          * :py:attr:`AutoCellRatio.DYNAMIC`, the ratio is determined from the
            :term:`active terminal` whenever :py:func:`get_cell_ratio` is called,
            though with some caching involved, such that the ratio is re-determined
            only if the terminal size changes.

    Raises:
        TypeError: An argument is of an inappropriate type.
        ValueError: An argument is of an appropriate type but has an
          unexpected/invalid value.
        term_image.exceptions.TermImageError: Auto cell ratio is not supported
          in the :term:`active terminal` or on the current platform.

    This value is taken into consideration when setting image sizes for **text-based**
    render styles, in order to preserve the aspect ratio of images drawn to the
    terminal.

    NOTE:
        Changing the cell ratio does not automatically affect any image that has a
        :term:`fixed size`. For a change in cell ratio to take effect, the image's
        size has to be re-set.

    ATTENTION:
        See :ref:`auto-cell-ratio` for details about the auto modes.
    """
    global _cell_ratio

    if isinstance(ratio, AutoCellRatio):
        if AutoCellRatio.is_supported is None:
            AutoCellRatio.is_supported = get_cell_size() is not None

        if not AutoCellRatio.is_supported:
            raise TermImageError(
                "Auto cell ratio is not supported in the active terminal or on the "
                "current platform"
            )
        elif ratio is AutoCellRatio.FIXED:
            # `(1, 2)` is a fallback in case the terminal doesn't respond in time
            _cell_ratio = truediv(*(get_cell_size() or (1, 2)))
        else:
            _cell_ratio = None
    elif isinstance(ratio, float):
        if ratio <= 0.0:
            raise ValueError(f"'ratio' must be greater than zero (got: {ratio})")
        _cell_ratio = ratio
    else:
        raise TypeError(
            "'ratio' must be a float or AutoCellRatio enum member "
            f"(got: {type(ratio).__name__})"
        )


class AutoCellRatio(Enum):
    """Values for setting :ref:`auto-cell-ratio`."""

    is_supported: Optional[bool]

    FIXED = auto()
    DYNAMIC = auto()


_cell_ratio = 0.5
AutoCellRatio.is_supported = None



from __future__ import annotations

__all__ = ("ITerm2Image",)

import io
import re
import sys
import warnings
from base64 import standard_b64encode
from operator import mul
from threading import Event
from typing import Any, Dict, Optional, Set, Tuple, Union

import PIL

from ..exceptions import TermImageWarning, _style_error
from ..utils import CSI, ESC, OSC, ST, get_terminal_name_version, get_terminal_size
from .common import GraphicsImage, ImageSource

# Constants for render methods
LINES = "lines"
WHOLE = "whole"


class ITerm2Image(GraphicsImage):
    """A render style using the iTerm2 inline image protocol.

    See :py:class:`GraphicsImage` for the complete description of the constructor.

    **Render Methods:**

    :py:class:`ITerm2Image` provides two methods of :term:`rendering` images, namely:

    LINES (default)
       Renders an image line-by-line i.e the image is evenly split across the number
       of lines it should occupy.

       Pros:

         * Good for use cases where it might be required to trim some lines of the
           image.

       Cons:

         * Image drawing is very slow on iTerm2 due to the terminal emulator's
           performance.

    WHOLE
       Renders an image all at once i.e the entire image data is encoded into one
       line of the :term:`rendered` output, such that the entire image is drawn once
       by the terminal and still occupies the correct amount of lines and columns.

       Pros:

         * Render results are more compact (i.e less in character count) than with
           the ``lines`` method since the entire image is encoded at once.
         * Image drawing is faster than with ``lines`` on most terminals.
         * Smoother animations.

       Cons:

          * This method currently doesn't work well on iTerm2 and WezTerm when the image
            height is greater than the terminal height.

    NOTE:
        The **LINES** method is the default only because it works properly in all cases,
        it's more advisable to use the **WHOLE** method except when the image height is
        greater than the terminal height or when trimming the image is required.

    The render method can be set with
    :py:meth:`set_render_method() <BaseImage.set_render_method>` using the names
    specified above.


    **Format Specification**

    See :ref:`format-spec`.

    ::

        [method] [ m {0 | 1} ] [ c {0-9} ]

    * ``method``: Render method override.

      Can be one of:

        * ``L``: **LINES** render method (current frame only, for animated images).
        * ``W``: **WHOLE** render method (current frame only, for animated images).
        * ``N``: Native animation. Ignored when used with non-animated images, WEBP
          images or ``ImageIterator``.

      Default: Current effective render method of the image.

    * ``m``: Cell content inter-mix policy (**Only supported in WezTerm**, ignored
      otherwise).

      * If the character after ``m`` is:

        * ``0``, contents of cells in the region covered by the image will be erased.
        * ``1``, the opposite, thereby allowing existing cell contents to show under
          transparent areas of the image.

      * If *absent*, defaults to ``m0``.
      * e.g ``m0``, ``m1``.

    * ``c``: ZLIB compression level, for images re-encoded in PNG format.

      * 1 -> best speed, 9 -> best compression, 0 -> no compression.
      * This results in a trade-off between render time and data size/draw speed.
      * If *absent*, defaults to ``c4``.
      * e.g ``c0``, ``c9``.


    ATTENTION:
        Currently supported terminal emulators include:

          * `iTerm2 <https://iterm2.com>`_
          * `Konsole <https://konsole.kde.org>`_ >= 22.04.0
          * `WezTerm <https://wezfurlong.org/wezterm/>`_
    """

    #: * ``x < 0``, JPEG encoding is disabled.
    #: * ``0 <= x <= 95``, JPEG encoding is used, with the specified quality, for
    #:   **most** non-transparent renders (at the cost of image quality).
    #:
    #: Only applies when not reading directly from file.
    #:
    #: By default, images are encoded in the PNG format (when not reading directly
    #: from file) but in some cases, higher compression might be desired.
    #: Also, JPEG encoding is significantly faster and can be useful to improve
    #: non-native animation performance.
    #:
    #: .. hint:: The transparency status of some images can not be correctly determined
    #:   in an efficient way at render time. To ensure the JPEG format is always used
    #:   for a re-encoded render, disable transparency or set a background color.
    JPEG_QUALITY: int = -1

    #: Maximum size (in bytes) of image data for native animation.
    #:
    #: | :py:class:`TermImageWarning<term_image.TermImageWarning>` is issued
    #:   (and shown **only the first time**, except a filter is set to do otherwise)
    #:   if the image data size for a native animation is above this value.
    #: | This value can be altered but should be done with caution to avoid excessive
    #:   memory usage.
    NATIVE_ANIM_MAXSIZE: int = 2 * 2**20  # 2 MiB

    #: * ``True``, image data is read directly from file when possible and no image
    #:   manipulation is required.
    #: * ``False``, images are always loaded and re-encoded, in the PNG format by
    #:   default.
    #:
    #: This is an optimization to reduce render times and is only applicable to the
    #: **WHOLE** render method, since the the **LINES** method inherently requires
    #: image manipulation.
    #:
    #: .. note:: This setting does not affect animations, native animations are always
    #:   read from file when possible and frames of non-native animations have to be
    #:   loaded and re-encoded.
    READ_FROM_FILE: bool = True

    _FORMAT_SPEC: Tuple[re.Pattern] = tuple(
        map(re.compile, "[LWN] m[01] c[0-9]".split(" "))
    )
    _render_methods: Set[str] = {LINES, WHOLE}
    _default_render_method: str = LINES
    _render_method: str = LINES
    _style_args = {
        "method": (
            None,
            (
                lambda x: isinstance(x, str),
                "Render method must be a string",
            ),
            (
                lambda x: x.lower() in __class__._render_methods,
                "Unknown render method for 'iterm2' render style",
            ),
        ),
        "mix": (
            False,
            (
                lambda x: isinstance(x, bool),
                "Cell content inter-mix policy must be a boolean",
            ),
            (lambda _: True, ""),
        ),
        "compress": (
            4,
            (
                lambda x: isinstance(x, int),
                "Compression level must be an integer",
            ),
            (
                lambda x: 0 <= x <= 9,
                "Compression level must be between 0 and 9, both inclusive",
            ),
        ),
        "native": (
            False,
            (
                lambda x: isinstance(x, bool),
                "Native animation policy must be a boolean",
            ),
            (lambda _: True, ""),
        ),
        "stall_native": (
            True,
            (
                lambda x: isinstance(x, bool),
                "Native animation execution policy must be a boolean",
            ),
            (lambda _: True, ""),
        ),
    }

    _TERM: str = ""
    _TERM_VERSION: str = ""

    @classmethod
    def clear(cls) -> None:
        """Clears all images on-screen.

        Required and works only on Konsole, as text doesn't overwrite images.
        """
        if cls._TERM == "konsole":
            # Seems Konsole utilizes the same image rendering implementation as it
            # uses for the kiity graphics protocol.
            _stdout_write(DELETE_ALL_IMAGES)

    def draw(
        self,
        *args,
        method: Optional[str] = None,
        mix: bool = False,
        compress: int = 4,
        native: bool = False,
        stall_native: bool = True,
        **kwargs,
    ):
        """Draws an image to standard output.

        Extends the common interface with style-specific parameters.

        Args:
            args: Positional arguments passed up the inheritance chain.
            method: Render method override. If ``None`` or not given, the current
              effective render method of the instance is used.
            mix: Cell content inter-mix policy (**Only supported in WezTerm**, ignored
              otherwise). If:

              * ``False``, contents of cells within the region covered by the image are
                erased.
              * ``True``, the opposite, thereby allowing existing text or image pixels
                to show under transparent areas of the image.

            compress: ZLIB compression level, for images re-encoded in PNG format.

              An integer between 0 and 9: 1 -> best speed, 9 -> best compression, 0 ->
              no compression. This results in a trade-off between render time and data
              size/draw speed.

            native: If ``True``, use native animation (if supported).

              * Ignored for non-animations.
              * *animate* must be ``True``.
              * *alpha*, *repeat*, *cached* and *style* do not apply.
              * Always loops infinitely.
              * No control over frame duration.
              * Not all animated image formats are supported e.g WEBP.
              * The limitations of the **WHOLE** render method also apply.
              * Normal restrictions for rendered/padding height of animations do not
                apply.

            stall_native: Native animation execution control. If:

              * ``True``, block until ``SIGINT`` (Ctrl+C) is recieved.
              * ``False``, return as soon as the image is transmitted.

            kwargs: Keyword arguments passed up the inheritance chain.

        Raises:
            term_image.exceptions.ITerm2ImageError: Native animation is not supported.

        See the ``draw()`` method of the parent classes for full details, including the
        description of other parameters.
        """
        if not (self._is_animated and kwargs.get("animate", True)):
            # Prevent the arguments from being passed on
            native = False
            stall_native = True

        arguments = locals()
        super().draw(
            *args,
            **kwargs,
            **{
                var: arguments[var]
                for var, default in __class__.draw.__kwdefaults__.items()
                if arguments[var] is not default
            },
        )

    @classmethod
    def is_supported(cls):
        if cls._supported is None:
            cls._supported = False

            name, version = get_terminal_name_version()
            if name in {"iterm2", "konsole", "wezterm"}:
                try:
                    if name != "konsole" or (
                        tuple(map(int, version.split("."))) >= (22, 4, 0)
                    ):
                        cls._supported = True
                        cls._TERM, cls._TERM_VERSION = name, version
                except ValueError:  # version string not "understood"
                    pass

        return cls._supported

    @classmethod
    def _check_style_format_spec(cls, spec: str, original: str) -> Dict[str, Any]:
        parent, (method, mix, compress) = cls._get_style_format_spec(spec, original)
        args = {}
        if parent:
            args.update(super()._check_style_format_spec(parent, original))
        if method == "N":
            args["native"] = True
        elif method:
            args["method"] = LINES if method == "L" else WHOLE
        if mix:
            args["mix"] = bool(int(mix[-1]))
        if compress:
            args["compress"] = int(compress[-1])

        return cls._check_style_args(args)

    def _display_animated(
        self,
        img,
        alpha,
        fmt,
        *args,
        mix: bool = False,
        native: bool = False,
        stall_native: bool = True,
        **kwargs,
    ):
        if native:
            if self._TERM == "konsole":
                raise _style_error(type(self))(
                    "Native animation is not supported in the active terminal"
                )
            if img.format == "WEBP":
                raise _style_error(type(self))("Native WEBP animation is not supported")
            try:
                print(
                    self._format_render(
                        self._render_image(img, alpha, mix=mix, native=True),
                        *fmt,
                    ),
                    end="",
                    flush=True,
                )
            except (KeyboardInterrupt, Exception):
                self._handle_interrupted_draw()
                raise
            else:
                stall_native and native_anim.wait()
        else:
            if not mix and self._TERM == "wezterm":
                lines = max(
                    (fmt or (None,))[-1] or get_terminal_size()[1] - self._v_allow,
                    self.rendered_height,
                )
                r_width = self.rendered_width
                erase_and_jump = f"{CSI}{r_width}X{CSI}{r_width}C"
                first_frame = self._format_render(
                    f"{erase_and_jump}\n" * (lines - 1) + f"{erase_and_jump}", *fmt
                )
                print(first_frame, f"\r{CSI}{lines - 1}A", sep="", end="", flush=True)

            super()._display_animated(img, alpha, fmt, *args, mix=True, **kwargs)

    @staticmethod
    def _handle_interrupted_draw():
        """Performs neccessary actions when image drawing is interrupted.

        If drawing is interruped while transmiting an image, it causes terminal to
        wait for more data (while consuming any output following) until the output
        reaches the expected payload size or ST (String Terminator) is written.
        """

        # End last transmission (does no harm if there wasn't an unterminated
        # transmission)
        # Konsole sometimes requires ST to be written twice.
        print(f"{ST * 2}", end="", flush=True)

    def _render_image(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        frame: bool = False,
        method: Optional[str] = None,
        mix: bool = False,
        compress: int = 4,
        native: bool = False,
    ) -> str:
        # Using `width=<columns>`, `height=<lines>` and `preserveAspectRatio=0` ensures
        # that an image always occupies the correct amount of columns and lines even if
        # the cell size has changed when it's drawn.
        # Since we use `width` and `height` control data keys, there's no need
        # upscaling an image on this end; ensures minimal payload.

        r_width, r_height = self.rendered_size

        # Workarounds
        is_on_konsole = self._TERM == "konsole"
        is_on_wezterm = self._TERM == "wezterm"
        jump_right = f"{CSI}{r_width}C"
        erase = f"{CSI}{r_width}X" if not mix and is_on_wezterm else ""

        file_is_readable = True
        if self._source_type is ImageSource.PIL_IMAGE:
            try:
                open(img.filename)
            except (AttributeError, OSError):
                file_is_readable = False

        if native and self._is_animated and not frame and img.format != "WEBP":
            if self._source_type is ImageSource.PIL_IMAGE:
                if file_is_readable:
                    compressed_image = open(img.filename, "rb")
                else:
                    try:
                        compressed_image = io.BytesIO()
                        img.save(compressed_image, img.format, save_all=True)
                    except ValueError:
                        raise _style_error(type(self))(
                            "Native animation not supported: This image was sourced "
                            "from a PIL image of an unknown format"
                        ) from None
            else:
                compressed_image = open(self._source, "rb")

            if img is not self._source:
                img.close()

            with compressed_image:
                compressed_image.seek(0, 2)
                if compressed_image.tell() > __class__.NATIVE_ANIM_MAXSIZE:
                    warnings.warn(
                        "Image data size above the maximum for native animation",
                        TermImageWarning,
                    )

                control_data = "".join(
                    (
                        f"size={compressed_image.tell()};width={r_width}"
                        f";height={r_height};preserveAspectRatio=0;inline=1:"
                    )
                )
                compressed_image.seek(0)
                return "".join(
                    (
                        f"{erase}{jump_right}\n" * (r_height - 1),
                        erase,
                        f"{CSI}{r_height - 1}A",
                        START,
                        control_data,
                        standard_b64encode(compressed_image.read()).decode(),
                        ST,
                    )
                )

        render_method = (method or self._render_method).lower()
        width, height = self._get_minimal_render_size(adjust=render_method == LINES)

        if (  # Read directly from file when possible and reasonable
            self.READ_FROM_FILE
            and not self._is_animated
            and file_is_readable
            and render_method == WHOLE
            and mul(*self._original_size) <= mul(*self._get_render_size())
            and (
                # None of the *alpha* options can affect these
                img.mode in {"1", "L", "RGB", "HSV", "CMYK"}
                # Alpha threshold is unused with graphics-based styles.
                # The transparency of some "P" mode images is missing on some terminals
                # Making the output inconsistent with other render styles.
                or (isinstance(alpha, float) and img.mode not in {"P", "PA"})
            )
        ):
            compressed_image = open(
                img.filename
                if self._source_type is ImageSource.PIL_IMAGE
                else self._source,
                "rb",
            )
            frame_img = None
        else:
            frame_img = img if frame else None
            img = self._get_render_data(
                img, alpha, size=(width, height), pixel_data=False, frame=frame
            )[0]  # fmt: skip
            if self.JPEG_QUALITY >= 0 and img.mode == "RGB":
                format = "jpeg"
                jpeg_quality = min(self.JPEG_QUALITY, 95)
            else:
                format = "png"
                jpeg_quality = None

            if render_method == LINES:
                raw_image = io.BytesIO(img.tobytes())
                compressed_image = io.BytesIO()
            else:
                compressed_image = io.BytesIO()
                img.save(
                    compressed_image,
                    format,
                    compress_level=compress,  # PNG
                    quality=jpeg_quality,
                )

        # clean up (ImageIterator uses one PIL image throughout)
        if frame_img is not img is not self._source:
            img.close()

        if render_method == LINES:
            # NOTE: It's more efficient to write separate strings to the buffer
            # separately than concatenate and write together.

            cell_height = height // r_height
            bytes_per_line = width * cell_height * (len(img.mode))
            control_data = (
                f";width={r_width};height=1;preserveAspectRatio=0;inline=1"
                f"{';doNotMoveCursor=1' * is_on_konsole}:"
            )

            with io.StringIO() as buffer, raw_image, compressed_image:
                for line in range(1, r_height + 1):
                    compressed_image.seek(0)
                    with PIL.Image.frombytes(
                        img.mode, (width, cell_height), raw_image.read(bytes_per_line)
                    ) as img:
                        img.save(
                            compressed_image,
                            format,
                            compress_level=compress,  # PNG
                            quality=jpeg_quality,
                        )
                    compressed_image.truncate()

                    buffer.write(erase)
                    buffer.write(f"{START}size={compressed_image.tell()}")
                    buffer.write(control_data)
                    buffer.write(
                        standard_b64encode(compressed_image.getvalue()).decode()
                    )
                    buffer.write(ST)
                    is_on_konsole and buffer.write(jump_right)
                    line < r_height and buffer.write("\n")

                return buffer.getvalue()

        # WHOLE
        with compressed_image:
            compressed_image.seek(0, 2)
            control_data = "".join(
                (
                    f"size={compressed_image.tell()};width={r_width}"
                    f";height={r_height};preserveAspectRatio=0;inline=1"
                    f"{';doNotMoveCursor=1' * is_on_konsole}:"
                )
            )
            compressed_image.seek(0)
            return "".join(
                (
                    "" if is_on_konsole else f"{erase}{jump_right}\n" * (r_height - 1),
                    erase,
                    "" if is_on_konsole else f"{CSI}{r_height - 1}A",
                    START,
                    control_data,
                    standard_b64encode(compressed_image.read()).decode(),
                    ST,
                    f"{jump_right}\n" * (r_height - 1) if is_on_konsole else "",
                    jump_right * is_on_konsole,
                )
            )


START = f"{OSC}1337;File="
DELETE_ALL_IMAGES = f"{ESC}_Ga=d;{ST}".encode()
native_anim = Event()
_stdout_write = sys.stdout.buffer.write



"""Main UI"""

from __future__ import annotations

import logging as _logging
import os
from operator import mul
from os.path import abspath, basename, islink
from pathlib import Path
from queue import Queue
from threading import Event
from typing import Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union

import PIL
import urwid

from .. import logging, notify, tui
from ..config import context_keys, expand_key
from .keys import (
    disable_actions,
    display_context_keys,
    enable_actions,
    keys,
    menu_nav,
    no_globals,
    set_image_grid_actions,
    set_image_view_actions,
    set_menu_actions,
    set_menu_count,
)
from .render import grid_render_queue
from .widgets import (
    Image,
    ImageCanvas,
    LineSquare,
    MenuEntry,
    image_box,
    image_grid,
    image_grid_box,
    info_bar,
    menu,
    placeholder,
    view,
    viewer,
)


def animate_image(image_w: Image, forced_render: bool = False) -> None:
    """Initializes an animation."""
    if not NO_ANIMATION and (
        mul(*image_w._ti_image._original_size) <= MAX_PIXELS or forced_render
    ):
        try:
            del image_w._ti_anim_finished
        except AttributeError:
            try:
                del image_w._ti_anim_ongoing
            except AttributeError:
                pass
        else:
            if image_w._ti_canv:  # Hasn't been removed
                image_w._ti_canv = None  # Deleting will break `ImageRenderManager`

        # Switched from this animated image earlier, to another animated image while
        # AnimRenderManager was waiting on a frame's duration
        # i.e the attributes didn't get to be removed by AnimRenderManager
        try:
            # If one fails, the rest shouldn't exist (removed by AnimRenderManager)
            del image_w._ti_frame
            del image_w._ti_force_render
            del image_w._ti_forced_anim_size_hash
        except AttributeError:
            pass

        # Only needs to be set once for an animation, not per frame
        if forced_render:
            image_w._ti_force_render = True


def display_images(
    dir: str,
    items: List[Tuple[str, Union[Image, type(...)]]],
    contents: Dict[str, Union[bool, Dict[str, Union[bool, dict]]]],
    prev_dir: str = "..",
    *,
    top_level: bool = False,
) -> Generator[None, int, bool]:
    """Displays images in _dir_ (and sub-directories, if '--recursive' is set)
    as yielded by ``scan_dir(dir)``.

    Args:
        - dir: Path to directory containing images.
        - items: An iterable of ``(entry, value)`` pairs, such as yielded by
          ``scan_dir(dir)`` i.e:

          - ``(str, Image)`` for images in *dir*, and
          - ``(str, Ellipsis)`` for sub-directories of *dir*

        - contents: Tree of directories containing readable images
          (such as returned by `check_dir(dir)`)
        - prev_dir: Path to set as working directory after displaying images in *dir*
          (default:  parent directory of *dir*)
        - top_level: Specifies if *dir* is the top level (For internal use only)

    Returns:
        The empty status of the (current) directory being exited i.e ``True`` if empty,
        otherwise ``False``.

    Receives:
        Via .send(), either:
          - a menu item position (-1 and above)
          - a flag denoting a certain action
    """
    global _grid_list, grid_path, last_non_empty_grid_path

    os.chdir(dir)

    # For `.tui.keys.set_menu_actions()`, called by `update_menu()`
    menu_is_complete = top_level or grid_scan_done.is_set()
    menu_scan_done.set() if menu_is_complete else menu_scan_done.clear()

    update_menu(items, top_level)
    next_menu.put((items, contents, menu_is_complete))

    entry = prev_pos = value = None  # Silence linter's `F821`
    pos = 0 if top_level else -1

    while True:
        if pos == -1:  # Cursor on top menu item ("..")
            if top_level:
                if items:
                    # Ensure ".." is not selectable at top level
                    # Possible when "Home" action is invoked
                    pos = 0
                    menu.focus_position = 1
                    continue
                else:
                    set_context("global")
            grid_active.clear()  # Grid not in view
            image_box._w.contents[1][0].contents[1] = (
                placeholder,
                ("weight", 1, False),
            )
            image_box.original_widget = placeholder  # halt image and anim rendering
            image_box.set_title("Image")
            view.original_widget = image_box
            getattr(ImageClass, "clear", lambda: True)()

        elif pos == OPEN:  # Implements "menu::Open" action (for non-image entries)
            if prev_pos == -1:
                # prev_pos can never be -1 at top level (See `pos == -1` branch above),
                # and even if it could `pos == BACK` still guards against it.
                pos = BACK
                continue

            # Ensure menu scanning is halted before WD is changed to prevent
            # `FileNotFoundError`s
            if not menu_scan_done.is_set():
                menu_acknowledge.clear()
                menu_change.set()
                menu_acknowledge.wait()
                menu_change.clear()

            # Ensure grid scanning is halted to avoid updating `_grid_list` which is
            # used as the next `menu_list` as is and to prevent `FileNotFoundError`s
            if not grid_scan_done.is_set():
                grid_acknowledge.clear()
                grid_active.clear()  # Grid not in view
                grid_acknowledge.wait()

            # To restore the menu on the way back
            menu_is_complete = menu_scan_done.is_set()

            logger.debug(f"Going into {abspath(entry)}/")
            empty = yield from display_images(
                entry,
                _grid_list,
                contents[entry],
                # Return to Top-Level Directory, OR
                # to the link's parent instead of the linked directory's parent
                os.getcwd() if top_level or islink(entry) else "..",
            )
            # Menu change already signaled by the BACK action from the exited directory

            # For `.tui.keys.set_menu_actions()`, called by `update_menu()`
            menu_scan_done.set() if menu_is_complete else menu_scan_done.clear()

            if empty:  # All entries in the exited directory have been deleted
                del items[prev_pos]
                del contents[entry]
                pos = min(prev_pos, len(items) - 1)
                # Restore the menu and view pane for the previous (this) directory,
                # while removing the empty directory entry.
                update_menu(items, top_level, pos)
                logger.debug(f"Removed empty directory entry '{entry}/' from the menu")
                notify.notify(f"Removed empty directory entry '{entry}/' from the menu")
            else:
                # Restore the menu and view pane for the previous (this) directory
                update_menu(items, top_level, prev_pos)
                pos = prev_pos

            next_menu.put((items, contents, menu_is_complete))
            continue  # Skip `yield`

        elif pos == BACK:  # Implements "menu::Back" action
            if not top_level:
                # Ensure menu scanning is halted before WD is changed to prevent
                # `FileNotFoundError`s
                if not menu_scan_done.is_set():
                    menu_acknowledge.clear()
                    menu_change.set()
                    menu_acknowledge.wait()
                    menu_change.clear()

                # Ensure grid scanning is halted before WD is changed to prevent
                # `FileNotFoundError`s
                if grid_active.is_set() and not grid_scan_done.is_set():
                    grid_acknowledge.clear()
                    grid_active.clear()  # Grid not in view
                    grid_acknowledge.wait()

                break

            # Since the execution context is not exited at the top-level, ensure pos
            # (and indirectly, prev_pos) always corresponds to a valid menu position.
            # By implication, this prevents an `IndexError` or rendering the wrong image
            # when coming out of a directory that was entered when prev_pos < -1.
            pos = prev_pos

        elif pos == DELETE:
            del items[prev_pos]
            pos = min(prev_pos, len(items) - 1)
            update_menu(items, top_level, pos)
            yield  # Displaying next image immediately will mess up confirmation overlay
            if DEBUG:
                info_bar.set_text(f"delete_pos={pos} {info_bar.text}")
            continue

        else:
            entry, value = items[pos]
            if isinstance(value, Image):
                grid_active.clear()  # Grid not in view
                image_box._w.contents[1][0].contents[1] = (value, ("weight", 1, False))
                image_box.set_title(entry)
                view.original_widget = image_box
                image_box.original_widget = value
                if value._ti_image._is_animated:
                    animate_image(value)
            else:  # Directory
                grid_acknowledge.clear()
                grid_active.set()  # Grid is in view

                next_grid.put((entry, contents[entry]))
                # No need to wait for acknowledgement since this is a new list instance
                _grid_list = []
                # Absolute paths work fine with symlinked images and directories,
                # as opposed to real paths, especially in path comparisons
                # e.g in `.tui.render.manage_grid_renders()`.
                grid_path = abspath(entry)

                if contents[entry].get("/") and grid_path != last_non_empty_grid_path:
                    grid_render_queue.put(None)  # Mark the start of a new grid
                    grid_change.set()
                    # Wait till GridRenderManager clears the cache
                    while grid_change.is_set():
                        pass
                    last_non_empty_grid_path = grid_path
                image_box.original_widget = placeholder  # halt image and anim rendering
                image_grid_box.set_title(grid_path + "/")
                view.original_widget = image_grid_box
                image_grid_box.base_widget._invalidate()

                if contents[entry].get("/"):
                    enable_actions("menu", "Switch Pane")
                else:
                    disable_actions("menu", "Switch Pane")

                # Wait for GridScanner to clear grid contents.
                # Not waiting could result in an `IndexError` raised by
                # `GridFlow.focus_position` in `GridFlow.generate_display_widget()`
                # when it meets the grid non-empty but it's then cleared by
                # GridScanner in the course of generating the display widget.
                grid_acknowledge.wait()

            getattr(ImageClass, "clear", lambda: True)() or ImageCanvas.change()

        prev_pos = pos
        pos = yield
        while pos == prev_pos:
            pos = yield
        if DEBUG:
            info_bar.set_text(f"pos={pos} {info_bar.text}")

    if not top_level:
        logger.debug(f"Going back to {abspath(prev_dir)}/")
        os.chdir(prev_dir)

    return menu_scan_done.is_set() and not items


def get_context() -> None:
    """Returns the current context"""
    return _context


def get_prev_context(n: int = 1) -> None:
    """Return the nth previous context (1 <= n <= 3)"""
    return _prev_contexts[n - 1]


def process_input(key: str) -> bool:
    if DEBUG:
        info_bar.set_text(f"{key!r} {info_bar.text}")

    found = False
    if key in keys["global"]:
        if (
            _context not in no_globals
            or _context == "global"
            or key in {"resized", expand_key[0]}
        ):
            func, state = keys["global"][key]
            func() if state else print("\a", end="", flush=True)
            found = True

    elif key[0] == "mouse press":  # strings also support subscription
        # change context if the pane in focus changed.
        if _context in {"image", "image-grid"} and viewer.focus_position == 0:
            set_context("menu")
            menu_nav()
            found = True
        elif _context == "menu":
            if viewer.focus_position == 1:
                if not context_keys["menu"]["Switch Pane"][4]:
                    # Set focus back to the menu if "menu::Switch Pane" is disabled
                    viewer.focus_position = 0
                else:
                    if view.original_widget is image_box:
                        set_context("image")
                        set_image_view_actions()
                    else:
                        set_context("image-grid")
                        set_image_grid_actions()
            else:  # Update image view
                menu_nav()
            found = True

    else:
        func, state = keys[_context].get(key, (None, None))
        if state:
            func()
        elif state is False:
            print("\a", end="", flush=True)
        found = state is not None

    return bool(found)


def scan_dir(
    dir: str,
    contents: Dict[str, Union[bool, Dict[str, Union[bool, dict]]]],
    last_entry: Optional[str] = None,
    sort_key: Optional[Callable] = None,
    *,
    notify_errors: bool = False,
) -> Generator[Tuple[str, Union[Image, type(...)]], None, int]:
    """Scans *dir* for readable images (and sub-directories containing such,
    if '--recursive' was set).

    Args:
        - dir: Path to directory to be scanned.
        - contents: Tree of directories containing readable images
          (as produced by ``.cli.check_dir(dir)``).
        - last_entry: The entry after which scanning should start, if ``None`` or
          not found, all entries in the directory are scanned.
        - sort_key: A callable to generate values to be used in sorting the directory
          entries.
        - notify_errors: If True, a notification showing the number of unreadable
          files will be displayed.

    Yields:
        A tuple ``(entry, value)``, where *entry* is ``str`` (the item name)
        and *value* is:

          - ``.tui.widgets.Image``, for images in *dir*, and
          - `Ellipsis` for sub-directories of *dir* (if '--recursive' is set).

    Returns:
        The number of unreadable files in *dir*.

    - If '--all' was set, hidden (.*) images and subdirectories are included.
    - Items are grouped into images and directories, sorted lexicographically and
      case-insensitively. Any preceding '.'s (dots) are ignored when sorting.
    - If a dotted entry has the same main-name as another entry, the dotted one comes
      first.
    """
    _entries = sorted(os.scandir(dir), key=sort_key or sort_key_lexi)
    entries = iter(_entries)
    if last_entry:
        for entry in entries:
            if entry.name == last_entry:
                break
        else:  # Start from the beginning if *last_entry* isn't found
            entries = _entries

    errors = 0
    for entry in entries:
        result = scan_dir_entry(entry, contents)
        if result == UNREADABLE:
            errors += 1
        yield result, (
            entry.name,
            (
                Image(ImageClass.from_file(entry.path))
                if result == IMAGE
                else ...
                if result == DIR
                else None
            ),
        )

    if notify_errors and errors:
        notify.notify(
            f"{errors} file(s) could not be read in {abspath(dir)!r}! Check the logs.",
            level=notify.ERROR,
        )

    return errors


def scan_dir_entry(
    entry: Union[os.DirEntry, Path],
    contents: Dict[str, Union[bool, Dict[str, Union[bool, dict]]]],
    entry_path: Optional[str] = None,
) -> int:
    """Scans a single directory entry and returns a flag indicating its kind."""
    if not SHOW_HIDDEN and entry.name.startswith("."):
        return HIDDEN
    if contents.get("/") and entry.is_file():
        try:
            PIL.Image.open(abspath(entry))
        except PIL.UnidentifiedImageError:
            # Reporting will apply to every non-image file :(
            return UNKNOWN
        except Exception:
            logging.log_exception(f"{abspath(entry)!r} could not be read", logger)
            return UNREADABLE
        else:
            return IMAGE
    if RECURSIVE and entry.name in contents:
        # `.cli.check_dir()` already eliminated bad symlinks
        return DIR

    return UNKNOWN


def scan_dir_grid() -> None:
    """Updates the image grid using ``scan_dir()``.

    This is designed to be executed in a separate thread, while certain grid details
    are passed in using the ``next_grid`` queue.

    For each valid entry, a tuple ``(entry, value)``, like in ``scan_dir()``,
    is appended to ``.tui.main._grid_list`` and adds the *value* to the
    grid widget (for image entries only), then updates the screen.
    """
    grid_contents = image_grid.contents
    while True:
        dir, contents = next_grid.get()
        grid_list = _grid_list
        grid_contents.clear()
        grid_acknowledge.set()  # Cleared grid contents
        grid_scan_done.clear()
        page_not_complete = True
        notify.start_loading()

        for result, item in scan_dir(dir, contents):
            if result == IMAGE:
                grid_list.append(item)
                grid_contents.append(
                    (
                        urwid.AttrMap(
                            LineSquare(item[1]), "unfocused box", "focused box"
                        ),
                        image_grid.options(),
                    )
                )
                image_grid_box.base_widget._invalidate()
                if page_not_complete:
                    if len(grid_contents) <= image_grid_box.base_widget._ti_page_ncell:
                        update_screen()
                    else:
                        page_not_complete = False
            elif result == DIR:
                grid_list.append(item)

            if not next_grid.empty():
                break
            if not grid_active.is_set():
                grid_acknowledge.set()
                break
        else:
            grid_scan_done.set()
            update_screen()
            # There is a possibility that `grid_scan_done` is read as "cleared"
            # in-between the end of the last iteration an here :)
            if not grid_active.is_set():
                grid_acknowledge.set()
        notify.stop_loading()


def scan_dir_menu() -> None:
    """Updates the menu list using ``scan_dir()``.

    This is designed to be executed in a separate thread, while certain menu details
    are passed in using the ``next_menu`` queue.

    For each valid entry, a tuple ``(entry, value)``, like in ``scan_dir()``,
    is appended to ``.tui.main.menu_list`` and appends a ``MenuEntry`` widget to the
    menu widget, then updates the screen.
    """
    menu_body = menu.body
    while True:
        items, contents, menu_is_complete = next_menu.get()
        if menu_is_complete:
            continue
        page_not_complete = True
        notify.start_loading()

        for result, item in scan_dir(
            ".", contents, items[-1][0] if items else None, notify_errors=True
        ):
            if result == IMAGE:
                items.append(item)
                menu_body.append(
                    urwid.AttrMap(
                        MenuEntry(item[0], "left", "clip"),
                        "default",
                        "focused entry",
                    )
                )
                if page_not_complete:
                    if len(items) <= menu._ti_height:
                        update_screen()
                    else:
                        page_not_complete = False
            elif result == DIR:
                items.append(item)
                menu_body.append(
                    urwid.AttrMap(
                        MenuEntry(item[0] + "/", "left", "clip"),
                        "default",
                        "focused entry",
                    )
                )
                if page_not_complete:
                    if len(items) <= menu._ti_height:
                        update_screen()
                    else:
                        page_not_complete = False

            if menu_change.is_set():
                menu_acknowledge.set()
                break
        else:
            menu_scan_done.set()
            set_menu_count()
            update_screen()
            # There is a possibility that `menu_scan_done` is read as "cleared"
            # in-between the end of the last iteration an here :)
            if menu_change.is_set():
                menu_acknowledge.set()
        notify.stop_loading()


def set_context(new_context) -> None:
    """Sets the current context and updates the Key/Action bar"""
    global _context

    if DEBUG:
        info_bar.set_text(f"{_prev_contexts} {info_bar.text}")
    _prev_contexts[1:] = _prev_contexts[:2]  # Right-shift older contexts
    _prev_contexts[0] = _context
    _context = new_context
    display_context_keys(new_context)
    if DEBUG:
        info_bar.set_text(f"{new_context!r} {_prev_contexts} {info_bar.text}")


def set_prev_context(n: int = 1) -> None:
    """Set the nth previous context as the current context (1 <= n <= 3)"""
    global _context

    if DEBUG:
        info_bar.set_text(f"{_prev_contexts} {info_bar.text}")
    _context = _prev_contexts[n - 1]
    display_context_keys(_context)
    _prev_contexts[:n] = []
    _prev_contexts.extend(["menu"] * n)
    if DEBUG:
        info_bar.set_text(f"{_prev_contexts} {info_bar.text}")


def sort_key_lexi(entry: Union[os.DirEntry, Path]):
    """Lexicographic ordering key function.

    Compatible with ``list.sort()``, ``sorted()``, etc.
    """
    name = entry.name
    return (
        chr(entry.is_file())  # group directories before files
        + name.lstrip(".").casefold()  # sorts within each group
        # '\0' makes the key for the non-hidden longer without affecting it's order
        # relative to other entries.
        + "\0" * (not name.startswith("."))  # hidden before non-hidden of the same name
    )


def update_menu(
    items: Iterable[Tuple[str, Union[Image, type(...)]]],
    top_level: bool = False,
    pos: int = -1,
) -> None:
    global menu_list, at_top_level
    menu_list, at_top_level = items, top_level

    menu.body[:] = [
        urwid.Text(("inactive", ".."))
        if top_level
        else urwid.AttrMap(MenuEntry(".."), "default", "focused entry")
    ] + [
        urwid.AttrMap(
            MenuEntry(
                (basename(entry) if at_top_level else entry) + "/" * (value is ...),
                "left",
                "clip",
            ),
            "default",
            "focused entry",
        )
        for entry, value in items
    ]
    menu.focus_position = pos + 1 + (at_top_level and pos == -1)
    set_menu_actions()
    set_menu_count()


def update_screen():
    """Triggers a screen redraw.

    Meant to be called from threads other than the thread in which the MainLoop is
    running.
    """
    try:
        os.write(update_pipe, b" ")
    except OSError as e:
        if e.errno != 9:
            logging.log_exception("Screen update failed", logger)


logger = _logging.getLogger(__name__)
quitting = Event()

# For grid scanning/display
grid_acknowledge = Event()
grid_active = Event()
grid_change = Event()
grid_scan_done = Event()
next_grid = Queue(1)

# The annotations below are put in comments for compatibility with Python 3.7
# as it doesn't allow names declared as `global` within functions to be annotated.

# # Set from within `display_images()`
_grid_list = None  #: Optional[list]
grid_path = None  #: Optional[str]
last_non_empty_grid_path = None  #: Optional[str]

# For menu scanning/listing
menu_acknowledge = Event()
menu_change = Event()
menu_scan_done = Event()
next_menu = Queue(1)

# For Context Management
_prev_contexts = ["menu"] * 3
_context = "menu"  # To avoid a NameError the first time set_context() is called.

# FLAGS for `display_images()`
OPEN = -2
BACK = -3
DELETE = -4

# FLAGS for `scan_dir*()`
UNKNOWN = -1
HIDDEN = 0
UNREADABLE = 1
IMAGE = 2
DIR = 3

# The annotations below are put in comments for compatibility with Python 3.7
# as it doesn't allow names declared as `global` within functions to be annotated.

# Set by `update_menu()`
menu_list = None  #: Optional[list]
at_top_level = None  #: Optional[bool]

# Set from `.tui.init()`
ImageClass: Optional[type] = None
displayer: Optional[Generator[None, int, bool]] = None
loop: Optional[tui.Loop] = None
update_pipe: Optional[int] = None

# # Corresponsing to command-line args
DEBUG: Optional[bool] = None
GRID_RENDERERS: Optional[int] = None
MAX_PIXELS: Optional[int] = None
NO_ANIMATION: Optional[bool] = None
RECURSIVE: Optional[bool] = None
SHOW_HIDDEN: Optional[bool] = None



"""
Utilities
=========
"""

from __future__ import annotations

__all__ = (
    "OS_IS_UNIX",
    "no_redecorate",
    "cached",
    "lock_tty",
    "unix_tty_only",
    "terminal_size_cached",
    "color",
    "get_cell_size",
    "get_fg_bg_colors",
    "get_terminal_size",
    "get_window_size",
    "query_terminal",
    "read_tty",
    "set_query_timeout",
    "write_tty",
)

import os
import re
import sys
import warnings
from array import array
from functools import wraps
from multiprocessing import Array, Process, Queue as mp_Queue, RLock as mp_RLock
from operator import floordiv
from pathlib import Path
from queue import Empty, Queue
from shutil import get_terminal_size as _get_terminal_size
from threading import RLock
from time import monotonic
from types import FunctionType
from typing import Callable, Optional, Tuple, Union

from .exceptions import TermImageWarning

# import logging

OS_IS_UNIX: bool
try:
    import fcntl
    import termios
    from select import select
except ImportError:
    OS_IS_UNIX = False
else:
    OS_IS_UNIX = True

#: If ``True``, :ref:`terminal-queries` are disabled, thereby affecting all
#: :ref:`dependent features <queried-features>`.
DISABLE_QUERIES: bool = False

#: A workaround for some terminal emulators (e.g older VTE-based ones) that wrongly
#: report window dimensions swapped.
#:
#: | If ``True``, the dimensions reported by the terminal emulator are swapped.
#: | This setting affects :ref:`auto-cell-ratio` computation.
SWAP_WIN_SIZE: bool = False

# Decorator Classes


class ClassInstanceMethod(classmethod):
    """A method which when invoked via the owner, behaves like a class method
    and when invoked via an instance, behaves like an instance method.
    """

    def __get__(self, obj, cls=None):
        # classmethod just uses cls directly if present.
        # Otherwise, type(obj) but we're not concerned with that.
        return super().__get__(None, obj or cls)


# Decorator Functions


def no_redecorate(decor: Callable) -> FunctionType:
    """Prevents a decorator from re-decorating objects.

    Args:
        decor: The decorator to be wrapped.
    """
    if getattr(decor, "_no_redecorate_", False):
        return decor

    @wraps(decor)
    def no_redecorate_wrapper(obj, *args, **kwargs):
        if not getattr(obj, f"_{decor.__name__}_", False):
            obj = decor(obj, *args, **kwargs)
            setattr(obj, f"_{decor.__name__}_", True)
        return obj

    no_redecorate_wrapper._no_redecorate_ = True
    return no_redecorate_wrapper


@no_redecorate
def cached(func: Callable) -> FunctionType:
    """Enables return value caching.

    Args:
        func: The function to be wrapped.

    An *_invalidate_cache* function is set as an attribute of the returned wrapper
    which when called clears the cache, so that the next call actually calls the
    wrapped function, no matter the value of *_cached*.

    NOTE:
        It's thread-safe, i.e there is no race condition between calls to the same
        decorated object across threads of the same process.

        Only works when function arguments, if any, are hashable.
    """

    @wraps(func)
    def cached_wrapper(*args, **kwargs):
        arguments = (args, tuple(kwargs.items()))
        with lock:
            try:
                return cache[arguments]
            except KeyError:
                return cache.setdefault(arguments, func(*args, **kwargs))

    def invalidate():
        with lock:
            cache.clear()

    cache = {}
    lock = RLock()
    cached_wrapper._invalidate_cache = invalidate

    return cached_wrapper


@no_redecorate
def lock_tty(func: Callable) -> FunctionType:
    """Synchronizes access to the :term:`active terminal`.

    Args:
        func: The function to be wrapped.

    When any decorated function is called, a re-entrant lock is acquired by the current
    process or thread and released after the call, such that any other decorated
    function called within another thread or subprocess has to wait till the lock is
    fully released (i.e has been released as many times as acquired) by the current
    process or thread.

    NOTE:
        | It automatically works across parent-/sub-processes, started directly or
          indirectly via ``multiprocessing.Process`` (or a subclass of it) and their
          threads.

    IMPORTANT:
        It only works if ``multiprocessing.synchronize`` is supported on the host
        platform.  If not supported, a warning is issued when starting a subprocess.
    """

    @wraps(func)
    def lock_tty_wrapper(*args, **kwargs):
        with _tty_lock:
            # logging.debug(f"{func.__name__} acquired TTY lock", stacklevel=3)
            return func(*args, **kwargs)

    return lock_tty_wrapper


@no_redecorate
def terminal_size_cached(func: Callable) -> FunctionType:
    """Enables return value caching based on the size of the :term:`active terminal`.

    Args:
        func: The function to be wrapped.

    If the terminal size is the same as for the last call, the last return value is
    returned. Otherwise the wrapped function is called and the new return value is
    stored and returned.

    An *_invalidate_terminal_size_cache* function is also set as an attribute of the
    returned wrapper which when called clears the cache, so that the next call actually
    calls the wrapped function.

    NOTE:
        It's thread-safe, i.e there is no race condition between calls to the same
        decorated callable across threads of the same process.
    """

    @wraps(func)
    def terminal_size_cached_wrapper(*args, **kwargs):
        with lock:
            ts = get_terminal_size()
            if not cache or ts != cache[1]:
                cache[:] = [func(*args, **kwargs), ts]
        return cache[0]

    def invalidate():
        with lock:
            cache.clear()

    cache = []
    lock = RLock()
    terminal_size_cached_wrapper._invalidate_terminal_size_cache = invalidate

    return terminal_size_cached_wrapper


@no_redecorate
def unix_tty_only(func: Callable) -> FunctionType:
    """Disable invokation of a function on a non-unix-like platform or when there is no
    :term:`active terminal`.

    Args:
        func: The function to be wrapped.
    """

    @wraps(func)
    def unix_only_wrapper(*args, **kwargs):
        return _tty and func(*args, **kwargs)

    unix_only_wrapper.__doc__ += """
    NOTE:
        Currently works on UNIX only, returns ``None`` on any other platform or when
        there is no :term:`active terminal`.
    """

    return unix_only_wrapper


# Non-decorators


def clear_queue(queue: Union[Queue, mp_Queue]):
    """Purges the given queue"""
    while True:
        try:
            queue.get(timeout=0.005)
        except Empty:
            break


def color(
    text: str, fg: Tuple[int] = (), bg: Tuple[int] = (), *, end: bool = False
) -> str:
    """Prepends *text* with 24-bit color escape codes for the given foreground and/or
    background RGB values, optionally ending with the color reset sequence.

    Args:
        text: String to be color-coded.
        fg: Foreground color.
        bg: Background color.
        end: If ``True``, the color reset sequence is appended to the returned string.

    Returns:
        The color-coded string.

    The color code is ommited for any of *fg* or *bg* that is empty.
    """
    return (FG_FMT * bool(fg) + BG_FMT * bool(bg) + "%s") % (
        *fg,
        *bg,
        text,
    ) + COLOR_RESET * end


@unix_tty_only
@terminal_size_cached
def get_cell_size() -> Optional[Tuple[int, int]]:
    """Returns the current size of a character cell in the :term:`active terminal`.

    Returns:
        The terminal cell size in pixels or `None` if undetermined.
    """
    ws = get_window_size()
    size = ws and tuple(map(floordiv, ws, get_terminal_size()))

    return size if size and 0 not in size else None


@cached
def get_fg_bg_colors(
    *, hex: bool = False
) -> Tuple[
    Union[None, str, Tuple[int, int, int]], Union[None, str, Tuple[int, int, int]]
]:
    """Returns the default FG and BG colors of the :term:`active terminal`.

    Returns:
        For each color:

        * an RGB 3-tuple, if *hex* is ``False``
        * an RGB hex string if *hex* is ``True``
        * ``None`` if undetermined
    """
    with _tty_lock:  # All of the terminal's reply isn't read in `query_terminal()`
        response = query_terminal(
            # Not all terminals (e.g VTE-based) support multiple queries in one escape
            # sequence, hence the repetition of OSC ... ST
            f"{OSC}10;?{ST}{OSC}11;?{ST}{CSI}c".encode(),
            # The response might contain a "c"; can't stop reading at "c"
            lambda s: not s.endswith(CSI.encode()),
        )
        if not DISABLE_QUERIES:
            read_tty()  # The rest of the response to `CSI c`

    fg = bg = None
    if response:
        for c, spec in RGB_SPEC.findall(response.decode().rpartition(ESC)[0]):
            if c == "10":
                fg = x_parse_color(spec)
            elif c == "11":
                bg = x_parse_color(spec)

    return tuple(
        rgb and ("#" + ("{:02x}" * 3).format(*rgb) if hex else rgb) for rgb in (fg, bg)
    )


@cached
def get_terminal_name_version() -> Tuple[Optional[str], Optional[str]]:
    """Queries the :term:`active terminal` for it's name and version"""
    with _tty_lock:  # the terminal's response to the query is not read all at once
        # Terminal name/version query + terminal attribute query
        # The latter is to speed up the entire query since most (if not all)
        # terminals should support it and most terminals treat queries as FIFO
        response = query_terminal(
            f"{CSI}>q{CSI}c".encode(),
            # The response might contain a "c"; can't stop reading at "c"
            lambda s: not s.endswith(CSI.encode()),
        )
        if not DISABLE_QUERIES:
            read_tty()  # The rest of the response to `CSI c`

    match = response and NAME_VERSION.fullmatch(response.decode().rpartition(ESC)[0])
    name, version = (
        match.groups()
        if match
        else map(os.environ.get, ("TERM_PROGRAM", "TERM_PROGRAM_VERSION"))
    )

    return (name and name.lower(), version)


def get_terminal_size() -> os.terminal_size:
    """Returns the current size of the :term:`active terminal`.

    Returns:
        The terminal size in columns and lines.

    Tries to query the :term:`active terminal` device and falls back to
    ``shutil.get_terminal_size()`` if that fails.

    This implementation still gives the correct size of the process' controlling
    terminal when output is redirected (in most cases), unlike the fallback.
    """
    if _tty:
        # faster and gives correct results when output is redirected
        try:
            size = os.get_terminal_size(_tty)
        except OSError:
            size = None
    else:
        size = None

    return size or _get_terminal_size()


@unix_tty_only
def get_window_size() -> Optional[Tuple[int, int]]:
    """Returns the current window size of the :term:`active terminal`.

    Returns:
        The terminal size in pixels.

    The speed of this implementation is almost entirely dependent on the terminal; the
    method it supports and its response time if it has to be queried.

    Returns ``None`` if the size couldn't be gotten in time or the terminal lacks
    support.
    """
    with _win_size_lock:
        ts = get_terminal_size()
        if ts == tuple(_win_size_cache[:2]):
            size = tuple(_win_size_cache[2:])
            return None if 0 in size else size

        # First try ioctl
        size = None
        buf = array("H", [0, 0, 0, 0])
        try:
            if not fcntl.ioctl(_tty, termios.TIOCGWINSZ, buf):
                size = tuple(buf[2:])
                if size == (0, 0):
                    size = None
        except OSError:
            pass

        if not size:
            # Then CSI 14 t
            # The second sequence is to speed up the entire query since most (if not
            # all) terminals should support it and most terminals treat queries as FIFO
            response = query_terminal(
                f"{CSI}14t{CSI}c".encode(), more=lambda s: not s.endswith(b"c")
            )
            size = (response or None) and WIN_SIZE.match(response.decode())
            if size:
                # XTWINOPS specifies (height, width)
                size = tuple(map(int, size.groups()))[::-1]

                # Termux seems to respond with (height / 2, width), though the values
                # are incorrect as they change with different zoom levels but still
                # always give a reasonable (almost always the same) cell size and ratio.
                if os.environ.get("SHELL", "").startswith("/data/data/com.termux/"):
                    size = (size[0], size[1] * 2)

        size = size[:: -SWAP_WIN_SIZE or 1] if size else (0, 0)
        _win_size_cache[:] = ts + size
        return None if 0 in size else size


def is_writable(path: Union[str, os.PathLike, Path]) -> bool:
    """Checks if a file path is writable or creatable.

    Returns:
      - ``True``, if:
        - the file exists and is writable
        - the file doesn't exists but can be created
      - ``False``, if:
        - the path points to a directory
        - the file exists but is unwritable
        - the file doesn't exists and cannot be created
    """
    path = Path(path).expanduser()
    writable = False

    if path.exists():
        if path.is_file() and os.access(path, os.W_OK):
            writable = True
    else:
        for path in path.parents:
            if path.exists():
                if path.is_dir() and os.access(path, os.W_OK):
                    writable = True
                break

    return writable


@unix_tty_only
@lock_tty
def query_terminal(
    request: bytes, more: Callable[[bytearray], bool], timeout: float = None
) -> Optional[bytes]:
    """Sends a query to the :term:`active terminal` and returns the response.

    Args:
        more: A callable, which when passed the response recieved so far, returns a
          boolean indicating if the response is incomplete or not. If it returns:

          * ``True``, more response is waited for.
          * ``False``, the recieved response is returned immediately.

        timeout: Time limit for awaiting a response from the terminal, in seconds
          (infinite if negative).

          If not given or ``None``, :py:data:`QUERY_TIMEOUT` (set by
          :py:func:`set_query_timeout`) is used.

    Returns:
        `None` if :py:data:`DISABLE_QUERIES` is true, else the terminal's response
        (empty, if no response is recieved after *timeout* is up).

    ATTENTION:
        Any unread input is discared before the query. If the input might be needed,
        it can be read using :py:func:`read_tty()` before calling this fucntion.
    """
    if DISABLE_QUERIES:
        return None

    old_attr = termios.tcgetattr(_tty)
    new_attr = termios.tcgetattr(_tty)
    new_attr[3] &= ~termios.ECHO  # Disable input echo
    try:
        termios.tcsetattr(_tty, termios.TCSAFLUSH, new_attr)
        write_tty(request)
        return read_tty(more, timeout or QUERY_TIMEOUT)
    finally:
        termios.tcsetattr(_tty, termios.TCSANOW, old_attr)


@unix_tty_only
@lock_tty
def read_tty(
    more: Callable[[bytearray], bool] = lambda _: True,
    timeout: Optional[float] = None,
    min: int = 0,
    *,
    echo: bool = False,
) -> Optional[bytes]:
    """Reads input directly from the :term:`active terminal` with/without blocking.

    Args:
        more: A callable, which when passed the input recieved so far, as a `bytearray`
          object, returns a boolean. If it returns:

          * ``True``, more input is waited for.
          * ``False``, the input recieved so far is returned immediately.

        timeout: Time limit for awaiting input, in seconds.
        min: Causes to block until at least the given number of bytes have been read.
        echo: If ``True``, any input while waiting is printed unto the screen.
          Any input before or after calling this function is not affected.

    Returns:
        The input read (empty, if *min* == ``0`` (default) and no input is recieved
        before *timeout* is up).

    If *timeout* is ``None`` (default), all available input is read without blocking.

    If *timeout* is not ``None`` and:

      * *timeout* < ``0``, it's infinite.
      * *min* > ``0``, input is waited for until at least *min* bytes have been read.

        After *min* bytes have been read, the following points apply with *timeout*
        being the leftover of the original *timeout*, if not yet used up.

      * *more* is not given, input is read or waited for until *timeout* is up.
      * *more* is given, input is read or waited for until ``more(input)`` returns
        ``False`` or *timeout* is up.

    Upon return or interruption, the :term:`active terminal` is **immediately** restored
    to the state in which it was met.
    """
    if not callable(more):
        raise TypeError("'more' must be callable")
    try:
        ret = more(bytearray())
    except Exception:
        raise ValueError(
            "'more' must be able to accept a single argument, a `bytearray` object"
        ) from None
    else:
        if not isinstance(ret, bool):
            raise TypeError("'more' must return a boolean")
    if not isinstance(timeout, (type(None), float)):
        raise TypeError("'timeout' must be `None` or a float")
    if not isinstance(min, int):
        raise TypeError("'min' must be an integer")
    if not isinstance(echo, bool):
        raise TypeError("'echo' must be a boolean")

    old_attr = termios.tcgetattr(_tty)
    new_attr = termios.tcgetattr(_tty)
    new_attr[3] &= ~termios.ICANON  # Disable cannonical mode
    new_attr[6][termios.VTIME] = 0  # Never block based on time
    if echo:
        new_attr[3] |= termios.ECHO  # Enable input echo
    else:
        new_attr[3] &= ~termios.ECHO  # Disable input echo
    # Block until *min* bytes are read, when *timeout* is not `None`.
    new_attr[6][termios.VMIN] = 0 if timeout is None else min

    input = bytearray()
    try:
        termios.tcsetattr(_tty, termios.TCSANOW, new_attr)
        r, w, x = [_tty], [], []

        if timeout is None:
            # VMIN=0 does not work as expected on some platforms when there's no input
            while select(r, w, x, 0.0)[0]:
                input.extend(os.read(_tty, 100))
        else:
            start = monotonic()
            if min > 0:
                input.extend(os.read(_tty, min))

                # Don't block based on based on amount of bytes anymore
                new_attr[6][termios.VMIN] = 0
                termios.tcsetattr(_tty, termios.TCSANOW, new_attr)

            duration = monotonic() - start
            while (timeout < 0 or duration < timeout) and more(input):
                # Reduces CPU usage
                # Also, VMIN=0 does not work on some platforms when there's no input
                if select(r, w, x, None if timeout < 0 else timeout - duration)[0]:
                    input.extend(os.read(_tty, 1))
                duration = monotonic() - start
            # logging.debug(duration)
    finally:
        termios.tcsetattr(_tty, termios.TCSANOW, old_attr)

    return bytes(input)


def set_query_timeout(timeout: float) -> None:
    """Sets the global timeout for :ref:`terminal-queries`.

    Args:
        timeout: Time limit for awaiting a response from the terminal, in seconds.

    Raises:
        TypeError: *timeout* is not a float.
        ValueError: *timeout* is less than or equal to zero.
    """
    global QUERY_TIMEOUT

    if not isinstance(timeout, float):
        raise TypeError(f"'timeout' must be a float (got: {type(timeout).__name__!r})")
    if timeout <= 0.0:
        raise ValueError(f"'timeout' must be greater than zero (got: {timeout!r})")

    QUERY_TIMEOUT = timeout


@unix_tty_only
@lock_tty
def write_tty(data: bytes) -> None:
    """Writes *data* to the :term:`active terminal` and waits until it's completely
    transmitted.
    """
    os.write(_tty, data)
    try:
        termios.tcdrain(_tty)
    except termios.error:  # "Permission denied" on some platforms e.g Termux
        pass


def x_parse_color(spec: str) -> Tuple[int, int, int]:
    """Converts an RGB device specification according to XParseColor"""
    # One hex char -> 4 bits
    return tuple(int(x, 16) * 255 // ((1 << (len(x) * 4)) - 1) for x in spec.split("/"))


def _process_start_wrapper(self, *args, **kwargs):
    global _tty_lock, _win_size_cache, _win_size_lock

    # Ensure it's not acquired by another process/thread before changing it.
    # The only way this can be countered is if the owner thread is the
    # one starting a process, which is very unlikely within a function meant
    # for input/output :|
    with _tty_lock:
        if isinstance(_tty_lock, type(RLock())):
            try:
                self._tty_lock = _tty_lock = mp_RLock()
            except ImportError:
                self._tty_lock = None
                warnings.warn(
                    "Multi-process synchronization is not supported on this platform!\n"
                    "Hence, if any subprocess will be writing/reading to/from the "
                    "active terminal, it may be unsafe to use any features requiring"
                    "terminal queries.\n"
                    "See https://term-image.readthedocs.io/en/stable/library/reference"
                    "/utils.html#terminal-queries\n"
                    "If any related issues occur, it's advisable to set "
                    "`term_image.utils.DISABLE_QUERIES = True`.\n"
                    "Simply set an 'ignore' filter for this warning (before starting "
                    "any subprocess) if not using any of the affected features.",
                    TermImageWarning,
                )
        else:
            self._tty_lock = _tty_lock

    with _win_size_lock:
        if isinstance(_win_size_lock, type(RLock())):
            try:
                self._win_size_cache = _win_size_cache = Array("i", _win_size_cache)
                _win_size_lock = _win_size_cache.get_lock()
            except ImportError:
                self._win_size_cache = None
        else:
            self._win_size_cache = _win_size_cache

    return _process_start_wrapper.__wrapped__(self, *args, **kwargs)


def _process_run_wrapper(self, *args, set_tty_lock: bool = True, **kwargs):
    global _tty_lock, _win_size_cache, _win_size_lock

    if set_tty_lock:
        if self._tty_lock:
            _tty_lock = self._tty_lock
        if self._win_size_cache:
            _win_size_cache = self._win_size_cache
            _win_size_lock = _win_size_cache.get_lock()

    return _process_run_wrapper.__wrapped__(self, *args, **kwargs)


QUERY_TIMEOUT = 0.1
RGB_SPEC = re.compile(r"\033](\d+);rgb:([\da-fA-F/]+)\033\\", re.ASCII)
WIN_SIZE = re.compile(r"\033\[4;(\d+);(\d+)t", re.ASCII)
NAME_VERSION = re.compile(r"\033P>\|(\w+)[( ]([^)\033]+)\)?\033\\", re.ASCII)

# Constants for escape sequences

ESC = "\033"
CSI = f"{ESC}["
OSC = f"{ESC}]"
ST = f"{ESC}\\"
BG_FMT = f"{CSI}48;2;%d;%d;%dm"
FG_FMT = f"{CSI}38;2;%d;%d;%dm"
COLOR_RESET = f"{CSI}m"

_tty: Optional[int] = None
_tty_lock = RLock()
_win_size_cache = [0] * 4
_win_size_lock = RLock()

if OS_IS_UNIX:
    for stream in ("out", "in", "err"):  # In order of priority
        try:
            _tty = os.ttyname(getattr(sys, f"__std{stream}__").fileno())
            break
        except (OSError, AttributeError):
            pass
    else:
        try:
            _tty = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        except OSError:
            warnings.warn(
                "It seems this process is not running within a terminal. "
                "Hence, some features will behave differently or be disabled.\n"
                "See https://term-image.readthedocs.io/en/stable/library/reference"
                "/utils.html#terminal-queries\n"
                "You can set an 'ignore' filter for this warning before loading "
                "`term_image`, if not using any of the features affected.",
                TermImageWarning,
            )
    if _tty:
        if isinstance(_tty, str):
            _tty = os.open(_tty, os.O_RDWR)

        Process.start = wraps(Process.start)(_process_start_wrapper)
        Process.run = wraps(Process.run)(_process_run_wrapper)

        # Shouldn't be needed since we're getting our own separate file descriptors
        # but the validity of the assumed safety is stil under probation
        """
        for name, value in vars(termios).items():
            if isinstance(value, BuiltinFunctionType):
                setattr(termios, name, lock_tty(value))
        """



"""
Custom Exceptions
=================
"""

from __future__ import annotations


class TermImageWarning(UserWarning):
    """Package-specific warning category."""


class TermImageError(Exception):
    """Exception baseclass. Raised for generic errors."""


class URLNotFoundError(FileNotFoundError, TermImageError):
    """Raised for 404 errors."""


class InvalidSizeError(ValueError, TermImageError):
    """Raised for invalid image sizes."""


# Style-specific exceptions


class StyleError(TermImageError):
    """Baseclass of style-specific exceptions.

    Never raised for errors pertaining to image classes defined in this package.
    Instead, the exception subclass specific to each image class is raised.

    Only raised for subclasses of :py:class:`BaseImage <term_image.image.BaseImage>`
    defined outside this package (which are not subclasses of any other image class
    defined in this package).

    Being the baseclass of all style-specific exceptions, it can be used be used to
    handle any style-specific error, regardless of the render style it originated from.
    """


class GraphicsImageError(StyleError):
    """Raised for errors specific to :py:class:`GraphicsImage
    <term_image.image.GraphicsImage>` and its subclasses defined outside this package.
    """


class TextImageError(StyleError):
    """Raised for errors specific to :py:class:`TextImage
    <term_image.image.TextImage>` and its subclasses defined outside this package.
    """


class BlockImageError(TextImageError):
    """Raised for errors specific to :py:class:`BlockImage
    <term_image.image.BlockImage>` and its subclasses defined outside this package.
    """


class ITerm2ImageError(GraphicsImageError):
    """Raised for errors specific to :py:class:`ITerm2Image
    <term_image.image.ITerm2Image>` and its subclasses defined outside this package.
    """


class KittyImageError(GraphicsImageError):
    """Raised for errors specific to :py:class:`KittyImage
    <term_image.image.KittyImage>` and its subclasses defined outside this package.
    """


__all__ = ["TermImageWarning"] + [
    name
    for name, obj in vars().items()
    if isinstance(obj, type) and issubclass(obj, TermImageError)
]
BaseImageError = StyleError  # Only to simplify `_style_error()`


def _style_error(cls: type):
    for cls in cls.__mro__:
        if cls.__module__.startswith("term_image.image"):
            return globals()[f"{cls.__name__}Error"]



"""Term-Image's Terminal User Interface"""

from __future__ import annotations

import argparse
import logging as _logging
import os
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, Tuple, Union

import urwid

from .. import logging
from ..utils import CSI, lock_tty, write_tty
from . import main, render
from .main import process_input, scan_dir_grid, scan_dir_menu, sort_key_lexi
from .widgets import Image, ImageCanvas, info_bar, main as main_widget


def init(
    args: argparse.Namespace,
    style_args: Dict[str, Any],
    images: Iterable[Tuple[str, Union[Image, Iterator]]],
    contents: dict,
    ImageClass: type,
) -> None:
    """Initializes the TUI"""
    global is_launched

    if args.debug:
        main_widget.contents.insert(
            -1, (urwid.AttrMap(urwid.Filler(info_bar), "input"), ("given", 1))
        )

    main.DEBUG = args.debug
    main.GRID_RENDERERS = args.grid_renderers
    main.MAX_PIXELS = args.max_pixels
    main.NO_ANIMATION = args.no_anim
    main.RECURSIVE = args.recursive
    main.SHOW_HIDDEN = args.all
    main.ImageClass = ImageClass
    main.loop = Loop(main_widget, palette, unhandled_input=process_input)
    main.update_pipe = main.loop.watch_pipe(lambda _: None)

    render.ANIM_CACHED = not args.cache_no_anim and (
        args.cache_all_anim or args.anim_cache
    )
    render.FRAME_DURATION = args.frame_duration
    render.REPEAT = args.repeat

    images.sort(
        key=lambda x: sort_key_lexi(
            Path(x[0] if x[1] is ... else x[1]._ti_image._source)
        )
    )
    main.displayer = main.display_images(".", images, contents, top_level=True)

    # `z_index=None` is pretty glitchy for animations with WHOLE method
    if ImageClass.style == "kitty" and ImageClass._KITTY_VERSION:
        render.anim_style_specs["kitty"] = "+L"
    for name in ("anim", "grid", "image"):
        specs = getattr(render, f"{name}_style_specs")
        if ImageClass.style == "kitty":
            # Kitty blends images at the same z-index
            if ImageClass._KITTY_VERSION:
                specs["kitty"] += "z"
            # Would've been removed if it had the default value
            if "compress" in style_args:
                specs["kitty"] += f"c{style_args['compress']}"
        elif ImageClass.style == "iterm2" and "compress" in style_args:
            specs["iterm2"] += f"c{style_args['compress']}"

    Image._ti_alpha = (
        "#"
        if args.no_alpha
        else (
            "#" + f"{args.alpha:f}"[1:]
            if args.alpha_bg is None
            else "#" + (args.alpha_bg or "#")
        )
    )
    Image._ti_grid_style_spec = render.grid_style_specs.get(ImageClass.style, "")

    # daemon, to avoid having to check if the main process has been interrupted
    menu_scanner = logging.Thread(target=scan_dir_menu, name="MenuScanner", daemon=True)
    grid_scanner = logging.Thread(target=scan_dir_grid, name="GridScanner", daemon=True)
    grid_render_manager = logging.Thread(
        target=render.manage_grid_renders,
        args=(args.grid_renderers,),
        name="GridRenderManager",
        daemon=True,
    )
    image_render_manager = logging.Thread(
        target=render.manage_image_renders,
        name="ImageRenderManager",
        daemon=True,
    )
    anim_render_manager = logging.Thread(
        target=render.manage_anim_renders,
        name="AnimRenderManager",
        daemon=True,
    )

    urwid.raw_display.Screen.get_available_raw_input = lock_tty(
        urwid.raw_display.Screen.get_available_raw_input
    )
    main.loop.screen.clear()
    main.loop.screen.set_terminal_properties(2**24)

    logger = _logging.getLogger(__name__)
    logging.log("Launching TUI", logger, direct=False)
    main.set_context("menu")
    is_launched = True
    menu_scanner.start()
    grid_scanner.start()
    grid_render_manager.start()
    image_render_manager.start()
    anim_render_manager.start()

    try:
        write_tty(f"{CSI}?1049h".encode())  # Switch to the alternate buffer
        next(main.displayer)
        main.loop.run()
        grid_render_manager.join()
        render.image_render_queue.put((None,) * 3)
        image_render_manager.join()
        render.anim_render_queue.put((None,) * 3)
        anim_render_manager.join()
        logging.log("Exited TUI normally", logger, direct=False)
    except Exception:
        main.quitting.set()
        render.image_render_queue.put((None,) * 3)
        image_render_manager.join()
        render.anim_render_queue.put((None,) * 3)
        anim_render_manager.join()
        raise
    finally:
        # urwid fails to restore the normal buffer on some terminals
        write_tty(f"{CSI}?1049l".encode())  # Switch back to the normal buffer
        main.displayer.close()
        is_launched = False
        os.close(main.update_pipe)


class Loop(urwid.MainLoop):
    def start(self):
        # Properly set expand key visbility at initialization
        self.unhandled_input("resized")
        return super().start()

    def process_input(self, keys):
        if "window resize" in keys:
            # Adjust bottom bar upon window resize
            keys.append("resized")
            getattr(main.ImageClass, "clear", lambda: True)() or ImageCanvas.change()
        return super().process_input(keys)


is_launched = False

palette = [
    ("default", "", "", "", "#ffffff", ""),
    ("default bold", "", "", "", "#ffffff, bold", ""),
    ("inactive", "", "", "", "#7f7f7f", ""),
    ("white on black", "", "", "", "#ffffff", "#000000"),
    ("black on white", "", "", "", "#000000", "#ffffff"),
    ("mine", "", "", "", "#ff00ff", "#ffff00"),
    ("focused entry", "", "", "", "standout", ""),
    ("unfocused box", "", "", "", "#7f7f7f", ""),
    ("focused box", "", "", "", "#ffffff", ""),
    ("green fg", "", "", "", "#00ff00", ""),
    ("red on green", "", "", "", "#ff0000,bold", "#00ff00"),
    ("key", "", "", "", "#ffffff", "#5588ff"),
    ("disabled key", "", "", "", "#7f7f7f", "#5588ff"),
    ("error", "", "", "", "bold", "#ff0000"),
    ("warning", "", "", "", "#ff0000,bold", ""),
    ("input", "", "", "", "standout", ""),
    ("notif context", "", "", "", "#0000ff,bold", ""),
]



from __future__ import annotations

__all__ = ("BlockImage",)

import io
import os
import warnings
from math import ceil
from operator import mul
from typing import Optional, Tuple, Union

import PIL

from ..utils import BG_FMT, COLOR_RESET, FG_FMT, get_fg_bg_colors
from .common import TextImage

warnings.filterwarnings("once", category=DeprecationWarning, module=__name__)

LOWER_PIXEL = "\u2584"  # lower-half block element
UPPER_PIXEL = "\u2580"  # upper-half block element


class BlockImage(TextImage):
    """A render style using unicode half blocks and ANSI 24-bit colour escape codes.

    See :py:class:`TextImage` for the description of the constructor.
    """

    @classmethod
    def is_supported(cls):
        if cls._supported is None:
            COLORTERM = os.environ.get("COLORTERM") or ""
            TERM = os.environ.get("TERM") or ""
            cls._supported = (
                "truecolor" in COLORTERM or "24bit" in COLORTERM or "256color" in TERM
            )

        return cls._supported

    def _get_render_size(self) -> Tuple[int, int]:
        return tuple(map(mul, self.rendered_size, (1, 2)))

    @staticmethod
    def _pixels_cols(
        *, pixels: Optional[int] = None, cols: Optional[int] = None
    ) -> int:
        return pixels if pixels is not None else cols

    @staticmethod
    def _pixels_lines(
        *, pixels: Optional[int] = None, lines: Optional[int] = None
    ) -> int:
        return ceil(pixels / 2) if pixels is not None else lines * 2

    def _render_image(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        frame: bool = False,
    ) -> str:
        # NOTE:
        # It's more efficient to write separate strings to the buffer separately
        # than concatenate and write together.

        def update_buffer():
            if alpha:
                no_alpha = False
                if a_cluster1 == 0 == a_cluster2:
                    buf_write(COLOR_RESET)
                    buf_write(" " * n)
                elif a_cluster1 == 0:  # up is transparent
                    buf_write(COLOR_RESET)
                    buf_write(FG_FMT % cluster2)
                    buf_write(LOWER_PIXEL * n)
                elif a_cluster2 == 0:  # down is transparent
                    buf_write(COLOR_RESET)
                    buf_write(FG_FMT % cluster1)
                    buf_write(UPPER_PIXEL * n)
                else:
                    no_alpha = True

            if not alpha or no_alpha:
                r, g, b = cluster2
                # Kitty does not render BG colors equal to the default BG color
                if is_on_kitty and cluster2 == bg_color:
                    r += r < 255 or -1
                buf_write(BG_FMT % (r, g, b))
                if cluster1 == cluster2:
                    buf_write(" " * n)
                else:
                    buf_write(FG_FMT % cluster1)
                    buf_write(UPPER_PIXEL * n)

        buffer = io.StringIO()
        buf_write = buffer.write  # Eliminate attribute resolution cost

        bg_color = get_fg_bg_colors()[1]
        is_on_kitty = self._is_on_kitty()

        width, height = self._get_render_size()
        frame_img = img if frame else None
        img, rgb, a = self._get_render_data(img, alpha, round_alpha=True, frame=frame)
        alpha = img.mode == "RGBA"

        # clean up (ImageIterator uses one PIL image throughout)
        if frame_img is not img is not self._source:
            img.close()

        rgb_pairs = (
            (
                zip(rgb[x : x + width], rgb[x + width : x + width * 2]),
                (rgb[x], rgb[x + width]),
            )
            for x in range(0, len(rgb), width * 2)
        )
        a_pairs = (
            (
                zip(a[x : x + width], a[x + width : x + width * 2]),
                (a[x], a[x + width]),
            )
            for x in range(0, len(a), width * 2)
        )

        row_no = 0
        # Two rows of pixels per line
        for (rgb_pair, (cluster1, cluster2)), (a_pair, (a_cluster1, a_cluster2)) in zip(
            rgb_pairs, a_pairs
        ):
            row_no += 2
            n = 0
            for (px1, px2), (a1, a2) in zip(rgb_pair, a_pair):
                # Color-code characters and write to buffer
                # when upper and/or lower pixel color/alpha-level changes
                if not (alpha and a1 == a_cluster1 == 0 == a_cluster2 == a2) and (
                    px1 != cluster1
                    or px2 != cluster2
                    or alpha
                    and (
                        # From non-transparent to transparent
                        a_cluster1 != a1 == 0
                        or a_cluster2 != a2 == 0
                        # From transparent to non-transparent
                        or 0 == a_cluster1 != a1
                        or 0 == a_cluster2 != a2
                    )
                ):
                    update_buffer()
                    cluster1 = px1
                    cluster2 = px2
                    if alpha:
                        a_cluster1 = a1
                        a_cluster2 = a2
                    n = 0
                n += 1
            # Rest of the line
            update_buffer()
            if row_no < height:  # last line not yet rendered
                buf_write(f"{COLOR_RESET}\n")

        buf_write(COLOR_RESET)  # Reset color after last line
        buffer.seek(0)  # Reset buffer pointer

        with buffer:
            return buffer.getvalue()



"""
.. Core Library Definitions
"""

from __future__ import annotations

__all__ = (
    "AutoImage",
    "from_file",
    "from_url",
    "ImageSource",
    "Size",
    "ImageMeta",
    "BaseImage",
    "GraphicsImage",
    "TextImage",
    "BlockImage",
    "ITerm2Image",
    "KittyImage",
    "ImageIterator",
)

from typing import Optional, Tuple, Union

import PIL

from .block import BlockImage  # noqa:F401
from .common import (  # noqa:F401
    BaseImage,
    GraphicsImage,
    ImageIterator,
    ImageMeta,
    ImageSource,
    Size,
    TextImage,
)
from .iterm2 import ITerm2Image  # noqa:F401
from .kitty import KittyImage  # noqa:F401


def AutoImage(
    image: PIL.Image.Image,
    *,
    width: Optional[int] = None,
    height: Optional[int] = None,
    scale: Tuple[float, float] = (1.0, 1.0),
) -> BaseImage:
    """Convenience function for creating an image instance from a PIL image instance.

    Returns:
        An instance of a subclass of :py:class:`BaseImage`.

    Same arguments and raised exceptions as the :py:class:`BaseImage` class constructor.
    """
    return _best_style()(image, width=width, height=height, scale=scale)


def from_file(
    filepath: str,
    **kwargs: Union[None, int, Tuple[float, float]],
) -> BaseImage:
    """Convenience function for creating an image instance from an image file.

    Returns:
        An instance of a subclass of :py:class:`BaseImage`.

    Same arguments and raised exceptions as :py:meth:`BaseImage.from_file`.
    """
    return _best_style().from_file(filepath, **kwargs)


def from_url(
    url: str,
    **kwargs: Union[None, int, Tuple[float, float]],
) -> BaseImage:
    """Convenience function for creating an image instance from an image URL.

    Returns:
        An instance of a subclass of :py:class:`BaseImage`.

    Same arguments and raised exceptions as :py:meth:`BaseImage.from_url`.
    """
    return _best_style().from_url(url, **kwargs)


def _best_style():
    for cls in _styles:
        if cls.is_supported():
            break
    return cls


# In order of preference, based on image quality and style performance/functionality.
# NOTE: 'iterm2' should come before 'kitty', if not any other reason, at least because
# the query for 'kitty' support detection messes up iTerm2's window title.
_styles = (ITerm2Image, KittyImage, BlockImage)



"""Event logging"""

from __future__ import annotations

import logging
import os
import sys
import warnings
from dataclasses import dataclass
from logging.handlers import RotatingFileHandler
from multiprocessing import Event as mp_Event
from threading import Thread
from typing import Optional, Set

from . import cli, notify


def init_log(
    logfile: str,
    level: int,
    debug: bool,
    multi: bool,
    quiet: bool,
    verbose: bool,
    verbose_log: bool,
) -> None:
    """Initialize application event logging"""
    global DEBUG, MULTI, QUIET, VERBOSE, VERBOSE_LOG

    logfile = os.path.expanduser(logfile)
    os.makedirs(os.path.dirname(logfile) or ".", exist_ok=True)
    handler = RotatingFileHandler(
        logfile,
        maxBytes=2**20,  # 1 MiB
        backupCount=1,
    )
    handler.addFilter(filter_)

    QUIET, VERBOSE, VERBOSE_LOG = quiet, verbose or debug, verbose_log
    DEBUG = debug = debug or level == logging.DEBUG
    if debug:
        level = logging.DEBUG
    elif VERBOSE or VERBOSE_LOG:
        level = logging.INFO

    FORMAT = (
        "({process}) ({asctime}) "
        + "{processName}: {threadName}: " * debug
        + "[{levelname}] {name}: "
        + "{funcName}: " * (debug and stacklevel_is_available)
        + "{message}"
    )
    logging.basicConfig(
        handlers=(handler,),
        format=FORMAT,
        style="{",
        level=level,
    )

    if debug:
        _logger.setLevel(logging.DEBUG)

    _logger.info("Starting a new session")
    _logger.info(f"Logging level set to {logging.getLevelName(level)}")

    if debug and not stacklevel_is_available:
        warnings.warn(
            "Please upgrade to Python 3.8 or later to get more detailed logs."
        )

    if not QUIET:
        notify.loading_indicator = Thread(target=notify.load, name="LoadingIndicator")
        notify.loading_indicator.start()

    if (
        not multi
        or cli.args.cli
        or (os.cpu_count() or 0) <= 2  # Avoid affecting overall system performance
        or sys.platform in {"win32", "cygwin"}
    ):
        MULTI = False
    else:
        try:
            import multiprocessing.synchronize  # noqa: F401
        except ImportError:
            MULTI = False
        else:
            MULTI = True

    if MULTI:
        from .logging_multi import process_multi_logs

        process_multi_logs.started = mp_Event()
        Thread(target=process_multi_logs, name="MultiLogger").start()
        process_multi_logs.started.wait()
        del process_multi_logs.started

        # Inherited by instances of `.logging_multi.Process`
        cli.interrupted = mp_Event()


def log(
    msg: str,
    logger: Optional[logging.Logger] = None,
    level: int = logging.INFO,
    context: str = "",
    *,
    direct: bool = True,
    file: bool = True,
    verbose: bool = False,
    loading: bool = False,
) -> None:
    """Report events to various destinations"""
    if loading:
        msg += "..."

    if verbose:
        if VERBOSE:
            logger.log(level, msg, **_kwargs)
            notify.notify(
                msg,
                getattr(notify, logging.getLevelName(level)),
                context,
                loading=loading,
            )
        elif VERBOSE_LOG:
            logger.log(level, msg, **_kwargs)
    else:
        if file:
            logger.log(level, msg, **_kwargs)
        if direct:
            notify.notify(
                msg,
                getattr(notify, logging.getLevelName(level)),
                context,
                loading=loading,
            )


def log_exception(
    msg: str,
    logger: logging.Logger,
    context: str = "",
    *,
    direct: bool = False,
    fatal: bool = False,
) -> None:
    """Report an error with the exception reponsible

    NOTE: Should be called from within an exception handler
    i.e from (also possibly in a nested context) within an except or finally clause.
    """
    if DEBUG:
        logger.exception(f"{msg} due to:", **_kwargs_exc)
    elif VERBOSE or VERBOSE_LOG:
        exc_type, exc, _ = sys.exc_info()
        logger.error(
            f"{msg} due to: ({exc_type.__module__}.{exc_type.__qualname__}) {exc}",
            **_kwargs,
        )
    else:
        logger.error(msg, **_kwargs)

    if VERBOSE and direct:
        notify.notify(msg, notify.CRITICAL if fatal else notify.ERROR, context)


# Not annotated because it's not directly used.
def _log_warning(msg, catg, fname, lineno, f=None, line=None):
    """Redirects warnings to the logging system.

    Intended to replace `warnings.showwarning()`.
    """
    _logger.warning(warnings.formatwarning(msg, catg, fname, lineno, line), **_kwargs)
    notify.notify(
        "Please view the logs for some warning(s).",
        level=notify.WARNING,
    )


# See "Filters" section in `logging` standard library documentation.
@dataclass
class Filter:
    disallowed: Set[str]

    def filter(self, record: logging.LogRecord) -> bool:
        return record.name.partition(".")[0] not in self.disallowed

    def add(self, name: str) -> None:
        self.disallowed.add(name)

    def remove(self, name: str) -> None:
        self.disallowed.remove(name)


class Thread(Thread):
    """A thread with integration into the logging system"""

    def __init__(self, *args, **kwargs):
        try:
            del kwargs["redirect_notifs"]
        except KeyError:
            pass
        super().__init__(*args, **kwargs)

    def run(self):
        _logger.debug("Starting")
        try:
            super().run()
        except Exception:
            log_exception(
                "Aborted" if logging.DEBUG else f"{self.name} was aborted", _logger
            )
        else:
            _logger.debug("Exiting")


filter_ = Filter({"PIL", "urllib3"})

# Writing to STDERR messes up output, especially with the TUI
warnings.showwarning = _log_warning

# Can't use "term_image", since the logger's level is changed.
# Otherwise, it would affect children of "term_image".
_logger = logging.getLogger("term-image")

# the _stacklevel_ parameter was added in Python 3.8
stacklevel_is_available = sys.version_info[:3] >= (3, 8, 0)
if stacklevel_is_available:
    # > log > logger.log > _log
    _kwargs = {"stacklevel": 2}
    # > exception-handler > log_exception > logger.exception > _log
    _kwargs_exc = {"stacklevel": 3}
else:
    _kwargs = _kwargs_exc = {}

# The annotations below are put in comments for compatibility with Python 3.7
# as it doesn't allow names declared as `global` within functions to be annotated.

# Set from within `init_log()`
DEBUG = None  #: Optional[bool]
MULTI = None  #: Optional[bool]
QUIET = None  #: Optional[bool]
VERBOSE = None  #: Optional[bool]
VERBOSE_LOG = None  #: Optional[bool]



"""Custom widget definitions and UI components assembly"""

from __future__ import annotations

import logging as _logging
from math import ceil
from operator import floordiv, mul, sub
from os.path import basename
from typing import List, Optional, Tuple

import urwid

from .. import logging
from ..config import config_options, expand_key, navi
from ..image import BaseImage, Size
from ..image.common import _ALPHA_THRESHOLD
from ..utils import get_terminal_size
from . import keys, main as tui_main
from .render import anim_render_queue, grid_render_queue, image_render_queue

# NOTE: Any new "private" attribute set on any subclass or instance of an urwid class
# should be prepended with "_ti" to prevent clashes with names used by urwid itself.


class GridListBox(urwid.ListBox):
    def __init__(self, grid: urwid.GridFlow):
        self._ti_grid = grid
        self._ti_ncell = 1
        self._ti_cell_width = grid.cell_width
        self._ti_grid_path = None
        self._ti_ncontent = 0
        self._ti_page_ncell = 1  # Used by GridScanner
        self._ti_topmost = None
        self._ti_top_trim = 0
        self._ti_next_index = 0

        return super().__init__([urwid.Divider()])

    def rows(self, size: Tuple[int, int], focus: bool = False) -> int:
        return self._ti_grid.rows(size[:1], focus)

    def keypress(self, size: Tuple[int, int], key: str) -> Optional[str]:
        if not size[1]:
            size = (size[0], 1)
        return super().keypress(size, key)

    def render(self, size: Tuple[int, int], focus: bool = False) -> urwid.Canvas:
        # 0, if maxcol < cell_width (maxcol = size[0]).
        # Otherwise, number of cells per row.
        ncell = sum(
            map(
                floordiv,
                # No of whole (cell_width + h_sep), columns left after last h_sep
                divmod(size[0], self._ti_grid.cell_width + self._ti_grid.h_sep),
                # if one cell_width can fit into the remaining space
                (1, self._ti_grid.cell_width),
            )
        )

        # The path takes care of "same directory"
        # The number of cells takes care of deletions in that directory.
        grid_path = tui_main.grid_path
        ncontent = len(self._ti_grid.contents)

        _row_pos = self.focus_position
        transfer_col_pos = False

        visible = self.calculate_visible(size)
        if visible[0]:
            (_, middle, *_), (top_trim, top), _ = visible
            topmost = top[-1][0] if top else middle
        else:
            topmost = top_trim = None

        if self._ti_grid_path == grid_path and not (
            self._ti_topmost is topmost and self._ti_top_trim == top_trim
        ):
            getattr(
                tui_main.ImageClass, "clear", lambda: True
            )() or ImageCanvas.change()

        self._ti_topmost = topmost
        self._ti_top_trim = top_trim

        if (
            self._ti_grid_path != grid_path  # Different grids
            or self._ti_ncontent != ncontent  # Different no of cells
            or not (ncell or self._ti_ncell)  # maxcol is and was < cell_width
            or ncell != self._ti_ncell  # Number of cells per row changed
            or self._ti_cell_width != self._ti_grid.cell_width  # cell_width changed
        ):
            # When maxcol < cell_width, the grid contents are not `Columns` widgets.
            # Instead, they're what would normally be the contents of the `Columns`.
            # If the grid is empty, then the `GridListBox` only contains a `Divider`

            # Old and new grids are both non-empty
            both_non_empty = self._ti_ncontent and ncontent
            # Conditions for transferring GridListBox's focus position
            transfer_row_pos = self._ti_grid_path == grid_path and both_non_empty

            if transfer_row_pos:
                # Conditions for transferring column focus position
                transfer_col_pos = ncell and self._ti_ncell
                # The 0-based index of the focused cell if the grid were laid out flat
                cell_index = (
                    # The GridListBox also contains dividers between columns
                    # i.e Column - Divider - Column - DIvider - Column - ...
                    # Hence the `// 2`
                    (self.focus_position // 2) * (self._ti_ncell or 1)
                    + (self._ti_ncell and self.focus.focus_position)
                )

            self._update_grid_contents(
                size[:1],
                ncell or 1,
                new=(
                    self._ti_grid_path != grid_path  # Different grids
                    or not (ncell or self._ti_ncell)  # maxcol is and was < cell_width
                    or ncell != self._ti_ncell  # Number of cells per row changed
                    # cell_width changed
                    or self._ti_cell_width != self._ti_grid.cell_width
                ),
            )

            if transfer_row_pos:
                # Ensure focus-position is not out-of-bounds
                # For the `* 2`, see the comments on cell_index calculation above
                self.focus_position = min(
                    len(self.body) - 1, cell_index // (ncell or 1) * 2
                )
            else:
                self.focus_position = 0

            if transfer_col_pos:
                # Ensure focus-position is not out-of-bounds
                col_pos = self.focus.focus_position = min(
                    len(self.focus.contents) - 1, cell_index % ncell
                )
            elif ncontent and ncell:
                self.focus.focus_position = 0

            if grid_path != self._ti_grid_path:
                # Maximum number of cells per grid page. Used by GridScanner
                self._ti_page_ncell = ncell * ceil(
                    size[1] / (ceil(self._ti_grid.cell_width / 2) + self._ti_grid.v_sep)
                )

            self._ti_grid_path = grid_path
            self._ti_ncontent = ncontent
            self._ti_ncell = ncell
            self._ti_cell_width = self._ti_grid.cell_width

        canv = super().render(size, focus)

        # For some reason, `GridListBox.render()` resets the focused column's
        # focus_position to 0 whenever its (the GridListBox's) own focus_position is
        # manually changed to a different position
        # So, the focus_position of the newly focused column has be set again after
        # `render()` and another render set in place
        if transfer_col_pos and _row_pos != self.focus_position:
            self.focus.focus_position = col_pos
            tui_main.update_screen()

        return canv

    def _update_grid_contents(
        self, size: Tuple[int, int], ncell: int, new: bool = True
    ) -> None:
        # The display widget is a `Divider` when the grid is empty
        if not self._ti_grid.contents:
            self._ti_next_index = 0
            self.body[:] = [urwid.Divider()]
            return

        if new:
            self._ti_next_index = 0
            self.body.clear()
        else:
            if self._ti_next_index:
                # Remove all cells after the previous *next_index* cos they are
                # officially just being added.
                # For the `* 2`, see the comments on cell_index calculation in
                # `render()` above.
                self.body[(self._ti_next_index // ncell) * 2 - 1 :] = [urwid.Divider()]
            else:
                self.body.clear()  # Remove the empty-grid Divider

        original = self._ti_grid._contents
        next_index = (len(original) // ncell) * ncell

        # Must include incomplete rows becaused the cells might've been counted with
        # *ncontent*.
        # They'll be removed before the next re-population if the grid wasn't complete
        # yet when *ncontent* was computed.
        # This way, the population can never be behind *ncontent*, ensuring the listbox
        # is always complete when the grid is complete.
        dummy = original[self._ti_next_index :]
        if not dummy:
            if not self._ti_next_index:
                # Would've been cleared earlier
                self.body[:] = [urwid.Divider()]
            return

        # Does not affect GridScanner as it uses a direct reference to the grid's
        # original contents list
        self._ti_grid._contents = urwid.MonitoredFocusList(dummy)

        self.body.extend(
            [
                content[0] if isinstance(content[0], urwid.Divider)
                # `.original_widget` gets rid of an unnecessary padding
                else content[0].original_widget
                for content in self._ti_grid.generate_display_widget(size).contents
            ]
        )

        self._ti_grid._contents = original
        self._ti_next_index = next_index


class Image(urwid.Widget):
    _sizing = frozenset(["box"])
    _selectable = True
    no_cache = ["render", "rows"]

    _ti_faulty_image = urwid.SolidFill("?")
    _ti_large_image = urwid.SolidFill("!")
    _ti_placeholder = urwid.SolidFill(".")

    _ti_force_render = False
    _ti_force_render_contexts = {"image", "full-image", "full-grid-image"}
    _ti_forced_anim_size_hash = None

    _ti_frame = None
    _ti_anim_ongoing = _ti_anim_finished = False

    _ti_faulty = False
    _ti_canv = None
    _ti_rendering = False

    _ti_grid_cache = {}

    # Updated from `.tui.init()`
    _ti_alpha = f"{_ALPHA_THRESHOLD}"[1:]
    _ti_grid_style_spec = ""

    def __init__(self, image: BaseImage):
        self._ti_image = image

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        return key

    def rows(self, size: Tuple[int, int], focus: bool = False) -> int:
        # Incompetent implementation due to the lack of *maxrows*
        return self._ti_image._valid_size(
            size[0],
            None,
            maxsize=get_terminal_size(),  # Omit 2-line allowance
        )[1]

    def render(self, size: Tuple[int, int], focus: bool = False) -> urwid.Canvas:
        context = tui_main.get_context()
        image = self._ti_image
        image.set_size(Size.AUTO, maxsize=size)

        # Forced render

        if mul(*image._original_size) > tui_main.MAX_PIXELS and not (
            self._ti_canv
            and (
                # will be resized later @ Rendering.
                self._ti_canv._ti_image_size == image.size
                # can either be SolidCanvas (faulty) or ImageCanvas
                if isinstance(self._ti_canv, ImageCanvas)
                # but faulty shouldn't be resized to allow re-rendering after resize
                else self._ti_canv.size == size
            )
            or self._ti_rendering
        ):
            if self._ti_force_render:
                # AnimRendermanager or `.tui.main.animate_image()` deletes
                # `_force_render` when the animation is done to avoid attribute
                # creation and deletion per frame
                if image._is_animated and not tui_main.NO_ANIMATION:
                    if not (self._ti_frame or self._ti_anim_finished):
                        self._ti_forced_anim_size_hash = hash(image.size)
                    elif hash(image.size) != self._ti_forced_anim_size_hash:
                        self._ti_force_render = False
                        if context in self._ti_force_render_contexts:
                            keys.enable_actions(context, "Force Render")
                        return __class__._ti_large_image.render(size, focus)
                else:
                    del self._ti_force_render
            else:
                if context in self._ti_force_render_contexts:
                    keys.enable_actions(context, "Force Render")
                return __class__._ti_large_image.render(size, focus)

        if context in self._ti_force_render_contexts:
            keys.disable_actions(context, "Force Render")

        # Grid cells

        if (
            view.original_widget is image_grid_box
            and context != "full-grid-image"
            # Grid render cell width adjusts when _maxcols_ < _cell_width_
            # `+2` cos `LineSquare` subtracts the columns for surrounding lines
            and size[0] + 2 == image_grid.cell_width
        ):
            canv = __class__._ti_grid_cache.get(basename(image._source))
            if not canv:
                grid_render_queue.put((image._source, size, self._ti_alpha))
                __class__._ti_grid_cache[basename(image._source)] = ...
                canv = __class__._ti_placeholder.render(size, focus)
            elif canv is ...:
                canv = __class__._ti_placeholder.render(size, focus)
            return canv

        # Rendering

        if view.original_widget is image_grid_box and context != "full-grid-image":
            # When the grid render cell width adjusts; when _maxcols_ < _cell_width_
            try:
                canv = ImageCanvas(
                    f"{image:1.1{self._ti_alpha}{self._ti_grid_style_spec}}"
                    .encode().split(b"\n"),  # fmt: skip
                    size,
                    image.rendered_size,
                )
            except Exception:
                canv = __class__._ti_faulty_image.render(size, focus)
        elif self._ti_frame:
            canv, repeat, frame_no = self._ti_frame
            if canv._ti_image_size != image.size:  # The canvas is always an ImageCanvas
                canv = (
                    placeholder
                    if (
                        # Workaround to erase text on wezterm without glitchy animation
                        tui_main.ImageClass.style == "iterm2"
                        and tui_main.ImageClass._TERM == "wezterm"
                    )
                    else __class__._ti_placeholder
                ).render(size)
                anim_render_queue.put(((repeat, frame_no), size, self._ti_force_render))
                self._ti_frame = None  # Avoid resending
                getattr(tui_main.ImageClass, "clear", lambda: True)()
            else:
                canv.size = size
        elif self._ti_canv and (
            self._ti_canv._ti_image_size == image.size
            # Can either be SolidCanvas (faulty) or ImageCanvas
            if isinstance(self._ti_canv, ImageCanvas)
            # but faulty shouldn't be resized to allow re-rendering after resize
            else self._ti_canv.size == size
        ):
            self._ti_canv.size = size
            canv = self._ti_canv
        else:
            if (
                image._is_animated
                and not tui_main.NO_ANIMATION
                and not self._ti_anim_finished
            ):
                if not self._ti_anim_ongoing:
                    anim_render_queue.put((self, size, self._ti_force_render))
                    self._ti_anim_ongoing = True
            elif not self._ti_rendering:
                self._ti_rendering = True
                image_render_queue.put((self, size, self._ti_alpha))
            canv = (
                placeholder
                if (
                    # Workaround to erase text on wezterm without glitchy animation
                    image._is_animated
                    and not tui_main.NO_ANIMATION
                    and tui_main.ImageClass.style == "iterm2"
                    and tui_main.ImageClass._TERM == "wezterm"
                )
                else __class__._ti_placeholder
            ).render(size)

        return canv


class ImageCanvas(urwid.Canvas):
    cacheable = False
    _ti_change_state = 0

    def __init__(
        self, lines: List[bytes], size: Tuple[int, int], image_size: Tuple[int, int]
    ):
        super().__init__()
        self.size = size
        self.lines = lines
        self._ti_image_size = image_size

    def cols(self) -> int:
        return self.size[0]

    def rows(self) -> int:
        return self.size[1]

    def content(self, trim_left=0, trim_top=0, cols=None, rows=None, attr_map=None):
        # In all our use cases, the canvas is never trimmed horizontally
        cols = self.cols()
        rows = rows or self.rows()  # Visble rows of the widget
        trim_bottom = self.rows() - trim_top - rows

        diff_x, diff_y = map(sub, self.size, self._ti_image_size)
        pad_up = diff_y // 2
        pad_down = diff_y - pad_up
        pad_left = diff_x // 2
        pad_right = diff_x - pad_left

        # If negative, they imply the number of lines to be trimmed off the image on
        # respective sides
        pad_up -= trim_top
        pad_down -= trim_bottom

        fill = b" " * cols
        fill_left = b" " * pad_left
        fill_right = b" " * pad_right + b"\b " * self._ti_change_state

        # Visible padding may be larger than the visible rows
        for _ in range(min(rows, pad_up)):
            yield [(None, "U", fill)]

        # See the description of `pad_up` and `pad_down` above
        for line in self.lines[-min(0, pad_up) : min(0, pad_down) or len(self.lines)]:
            yield [(None, "U", fill_left + line + fill_right)]

        # Visible padding may be larger than the visible rows
        for _ in range(min(rows, pad_down)):
            yield [(None, "U", fill)]

    @classmethod
    def change(cls):
        """Causes the canvas to embed or not embed some hidden text on every line of
        the image, such that every line of the image is seen as different in each state.

        ``urwid`` will not redraw lines that have not changed since the last redaw.
        So this is to trick ``urwid`` into taking every line containing a part of an
        image as different in each state.

        This is used to force redraws of all images on screen, particularly when their
        positions do not change much e.g when images need to be cleared in kitty.
        """
        cls._ti_change_state = (cls._ti_change_state + 1) % 3


class LineSquare(urwid.LineBox):
    no_cache = ["render", "rows"]

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        # Prevents `Image.rows()` from being called,
        # in order to get the correct no of rows for a `<LinesSquare <Image>>` widget
        original_middle = self._wrapped_widget.contents[1]
        new_middle = LineSquareMiddleColumns(
            [x[0] for x in original_middle[0].contents],
            box_columns=(0, 2),
            focus_column=original_middle[0].focus_position,
        )
        new_middle.contents[0] = (new_middle.contents[0][0], ("given", 1, True))
        new_middle.contents[2] = (new_middle.contents[2][0], ("given", 1, True))
        self._wrapped_widget.contents[1] = (new_middle, original_middle[1])

    def rows(self, size: Tuple[int, int], focus: bool = False) -> int:
        return ceil(size[0] / 2)

    def render(self, size: Tuple[int, int], focus: bool = False) -> urwid.Canvas:
        return super().render((size[0], ceil(size[0] / 2)), focus)


# To prevent `Image.rows()` from being called,
# in order to get the correct no of rows for a `<LinesSquare <Image>>` widget
class LineSquareMiddleColumns(urwid.Columns):
    no_cache = ["render", "rows"]

    def rows(self, size: Tuple[int, int], focus: bool = False) -> int:
        return ceil(size[0] / 2) - 2


class MenuEntry(urwid.Text):
    _selectable = True

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        return key


class MenuListBox(urwid.ListBox):
    def keypress(self, size: Tuple[int, int], key: str) -> Optional[str]:
        if not size[1]:
            size = (size[0], 1)
        ret = super().keypress(size, key)
        return key if key in navi else ret

    def render(self, size: Tuple[int, int], focus: bool = False):
        self._ti_height = size[1]  # Used by MenuScanner
        return super().render(size, focus)


class NoSwitchColumns(urwid.Columns):
    _command_map = urwid.command_map.copy()
    _command_map._command.clear()


class PlaceHolder(urwid.SolidFill):
    _selectable = True  # Prevents _image_box_ from being completely un-selectable

    def keypress(self, size: Tuple[int, int], key: str) -> str:
        return key


logger = _logging.getLogger(__name__)

placeholder = PlaceHolder(" ")
menu = MenuListBox(urwid.SimpleFocusListWalker([]))
menu_box = urwid.LineBox(menu, "List", "left")
image_grid = urwid.GridFlow([placeholder], config_options.cell_width, 2, 1, "left")
image_box = urwid.LineBox(placeholder, "Image", "left")
image_grid_box = urwid.LineBox(urwid.Padding(GridListBox(image_grid)), "Image", "left")
view = urwid.AttrMap(image_box, "unfocused box", "focused box")
viewer = NoSwitchColumns(
    [(20, urwid.AttrMap(menu_box, "unfocused box", "focused box")), view]
)
banner = urwid.LineBox(
    urwid.AttrMap(
        urwid.Filler(urwid.Text(("red on green", "Term-Image"), "center")),
        "red on green",
    ),
)
loading = urwid.Text("", "center")
notifications = urwid.Pile([])
notif_bar = urwid.Columns([(3, urwid.Filler(loading)), urwid.Filler(notifications)])
pile = urwid.Pile([(3, banner), viewer], 1)

info_bar = urwid.Text("")
key_bar = urwid.Filler(urwid.Text([[("mine", "cool"), " "]] * 19 + [("mine", "cool")]))
expand = urwid.Filler(urwid.Text(f"\u25B2 [{expand_key[1]}]", align="right"), "middle")
bottom_bar = urwid.Columns([key_bar, (len(expand.original_widget.text), expand)], 2)

main = urwid.Pile([pile, (1, bottom_bar)], 0)

confirmation = urwid.Text("", "center")
confirmation_overlay = urwid.Overlay(
    urwid.LineBox(
        urwid.Filler(confirmation),
        "",
        "center",
        None,
        "\u2554",
        "\u2550",
        "\u2551",
        "\u2557",
        "\u255a",
        "\u2551",
        "\u2550",
        "\u255d",
    ),
    placeholder,
    "center",
    ("relative", 25),
    "middle",
    ("relative", 25),
    50,
    3,
)

overlay = urwid.Overlay(
    urwid.LineBox(
        urwid.ListBox([placeholder]),
        "Help",
        "center",
        "default bold",
        "\u2554",
        "\u2550",
        "\u2551",
        "\u2557",
        "\u255a",
        "\u2551",
        "\u2550",
        "\u255d",
    ),
    placeholder,
    "center",
    ("relative", 50),
    "middle",
    ("relative", 75),
    100,
    5,
)



"""Definitions for deferred image rendering"""

from __future__ import annotations

import logging as _logging
from multiprocessing import Event as mp_Event, Queue as mp_Queue
from os.path import split
from queue import Empty, Queue
from threading import Event
from typing import Optional, Union

from .. import logging, notify
from ..image import Size
from ..logging_multi import Process
from ..utils import clear_queue


def manage_anim_renders() -> None:
    from .main import ImageClass, update_screen
    from .widgets import ImageCanvas, image_box

    def next_frame() -> bool:
        frame, repeat, frame_no, size, rendered_size = frame_render_out.get()
        if not_skip() and (not forced or image_w._ti_force_render):
            if frame:
                canv = ImageCanvas(frame.encode().split(b"\n"), size, rendered_size)
                image_w._ti_image._seek_position = frame_no
                image_w._ti_frame = (canv, repeat, frame_no)
            else:
                image_w._ti_anim_finished = True
                image_w._ti_image.seek(0)
        # If this image is the one currently displayed, it's either:
        # - forced but size changed -> End animation; Remove attributes
        # - a size change -> Continue animation; Do not remove attributes
        # - a restart (moved to another entry and back) -> Animation will be ended at
        #   restart; attributes already removed in `.main.animate_image()`
        elif image_w is not image_box.original_widget or forced:
            frame_render_in.put((..., None, None))
            clear_queue(frame_render_out)  # In case output is full
            frame = None

        if not frame:
            try:
                # If one fails, the rest shouldn't exist (removed in `animate_image()`)
                del image_w._ti_anim_ongoing
                del image_w._ti_frame
                if forced:
                    # See "Forced render" section of `.widgets.Image.render()`
                    del image_w._ti_force_render
                    del image_w._ti_forced_anim_size_hash
            except AttributeError:
                pass

        update_screen()
        return bool(frame)

    def not_skip():
        return image_w is image_box.original_widget and anim_render_queue.empty()

    frame_render_in = (mp_Queue if logging.MULTI else Queue)()
    frame_render_out = (mp_Queue if logging.MULTI else Queue)(20)
    ready = (mp_Event if logging.MULTI else Event)()
    renderer = (Process if logging.MULTI else logging.Thread)(
        target=render_frames,
        args=(
            frame_render_in,
            frame_render_out,
            ready,
            ImageClass,
            anim_style_specs.get(ImageClass.style, ""),
            REPEAT,
            ANIM_CACHED,
        ),
        name="FrameRenderer",
        redirect_notifs=True,
    )
    renderer.start()

    frame_duration = None
    image_w = None  # Silence flake8's F821

    try:
        while True:
            try:
                data, size, forced = anim_render_queue.get(timeout=frame_duration)
            except Empty:
                if not next_frame():
                    frame_duration = None
            else:
                if not data:
                    break

                notify.start_loading()

                if anim_render_queue.empty():
                    ready.clear()
                    frame_render_in.put((..., None, None))
                    clear_queue(frame_render_out)  # In case output is full
                    ready.wait()
                    # multiprocessing queues are not so reliable
                    clear_queue(frame_render_out)

                if isinstance(data, tuple):
                    if not_skip():
                        frame_render_in.put((data, size, image_w._ti_alpha))
                        if not next_frame():
                            frame_duration = None
                    elif image_w is not image_box.original_widget:
                        # The next item in the queue is NOT a size change
                        frame_duration = None
                else:
                    # Safe, since the next item in the queue cannot be a size change
                    # cos no animation is ongoing
                    frame_duration = None

                    image_w = data
                    if not_skip():
                        frame_render_in.put(
                            (image_w._ti_image._source, size, image_w._ti_alpha)
                        )
                        # Ensures successful deletion if the displayed image has
                        # changed before the first frame is ready
                        image_w._ti_frame = None

                        if next_frame():
                            frame_duration = (
                                FRAME_DURATION or image_w._ti_image._frame_duration
                            )

                notify.stop_loading()
    finally:
        clear_queue(frame_render_in)
        frame_render_in.put((None,) * 3)
        clear_queue(frame_render_out)  # In case the renderer is blocking on `put()`
        renderer.join()
        clear_queue(anim_render_queue)


def manage_image_renders():
    from .main import ImageClass, update_screen
    from .widgets import Image, ImageCanvas, image_box

    def not_skip():
        # If this image is the one currently displayed but the queue is non empty,
        # it means some "forth and back" has occured.
        # Skipping this render avoids the possibility of wasting time with this render
        # in the case where the image size has changed.
        return image_w is image_box.original_widget and image_render_queue.empty()

    multi = logging.MULTI
    image_render_in = (mp_Queue if multi else Queue)()
    image_render_out = (mp_Queue if multi else Queue)()
    renderer = (Process if multi else logging.Thread)(
        target=render_images,
        args=(
            image_render_in,
            image_render_out,
            ImageClass,
            image_style_specs.get(ImageClass.style, ""),
        ),
        kwargs=dict(out_extras=False, log_faults=True),
        name="ImageRenderer",
        redirect_notifs=True,
    )
    renderer.start()

    faulty_image = Image._ti_faulty_image
    last_image_w = image_box.original_widget
    # To prevent an `AttributeError` with the first deletion, while avoiding `hasattr()`
    last_image_w._ti_canv = None

    try:
        while True:
            # A redraw is neccesary even when the render is skipped, in case the skipped
            # render is of the currently displayed image.
            # So that a new render can be sent in (after `._ti_rendering` is unset).
            # Otherwise, the image will remain unrendered until a redraw.
            update_screen()

            image_w, size, alpha = image_render_queue.get()
            if not image_w:
                break

            if not not_skip():
                del image_w._ti_rendering
                continue

            image_render_in.put(
                (
                    image_w._ti_image._source,
                    size,
                    alpha,
                    image_w._ti_faulty,
                )
            )
            notify.start_loading()
            render, rendered_size = image_render_out.get()

            if not_skip():
                del last_image_w._ti_canv
                if render:
                    image_w._ti_canv = ImageCanvas(
                        render.encode().split(b"\n"), size, rendered_size
                    )
                else:
                    image_w._ti_canv = faulty_image.render(size)
                    # Ensures a fault is logged only once per `Image` instance
                    if not image_w._ti_faulty:
                        image_w._ti_faulty = True
                last_image_w = image_w

            del image_w._ti_rendering
            notify.stop_loading()
    finally:
        clear_queue(image_render_in)
        image_render_in.put((None,) * 4)
        renderer.join()
        clear_queue(image_render_queue)


def manage_grid_renders(n_renderers: int):
    """Manages grid cell rendering.

    Intended to be executed in a separate thread of the main process.

    If multiprocessing is enabled and *n_renderers* > 0, it spwans *n_renderers*
    subprocesses to render the cells and handles their proper termination.
    Otherwise, it starts a single new thread to render the cells.
    """
    from . import main
    from .main import ImageClass, grid_active, grid_change, quitting, update_screen
    from .widgets import Image, ImageCanvas, image_grid

    multi = logging.MULTI and n_renderers > 0
    grid_render_in = (mp_Queue if multi else Queue)()
    grid_render_out = (mp_Queue if multi else Queue)()
    renderers = [
        (Process if multi else logging.Thread)(
            target=render_images,
            args=(
                grid_render_in,
                grid_render_out,
                ImageClass,
                grid_style_specs.get(ImageClass.style, ""),
            ),
            kwargs=dict(out_extras=True, log_faults=False),
            name="GridRenderer" + f"-{n}" * multi,
            redirect_notifs=True,
        )
        for n in range(n_renderers if multi else 1)
    ]
    for renderer in renderers:
        renderer.start()

    cell_width = grid_path = None  # Silence flake8's F821
    faulty_image = Image._ti_faulty_image
    grid_cache = Image._ti_grid_cache
    new_grid = False

    try:
        while True:
            while not (
                grid_active.wait(0.1)
                or quitting.is_set()
                or not grid_render_out.empty()
            ):
                pass
            if quitting.is_set():
                break

            if new_grid or grid_change.is_set():  # New grid
                grid_cache.clear()
                grid_change.clear()  # Signal "cache cleared"
                if not new_grid:  # The starting `None` hasn't been gotten
                    while grid_render_queue.get():
                        pass
                for q in (grid_render_in, grid_render_out):
                    while True:
                        try:
                            q.get(timeout=0.005)
                            notify.stop_loading()
                        except Empty:
                            break
                cell_width = image_grid.cell_width
                grid_path = main.grid_path
                new_grid = False

            if grid_change.is_set():
                continue

            if grid_active.is_set():
                try:
                    image_info = grid_render_queue.get(timeout=0.02)
                except Empty:
                    pass
                else:
                    if not image_info:  # Start of a new grid
                        new_grid = True
                        continue
                    grid_render_in.put(image_info)
                    notify.start_loading()

            if grid_change.is_set():
                continue

            try:
                image_path, image, size, rendered_size = grid_render_out.get(
                    timeout=0.02
                )
            except Empty:
                pass
            else:
                dir, entry = split(image_path)
                # The directory and cell-width checks are to filter out any remnants
                # that were still being rendered at the other end
                if (
                    not grid_change.is_set()
                    and dir == grid_path
                    and size[0] + 2 == cell_width
                ):
                    grid_cache[entry] = (
                        ImageCanvas(image.encode().split(b"\n"), size, rendered_size)
                        if image
                        else faulty_image.render(size)
                    )
                    if grid_active.is_set():
                        update_screen()
                notify.stop_loading()
    finally:
        clear_queue(grid_render_in)
        for renderer in renderers:
            grid_render_in.put((None,) * 3)
        for renderer in renderers:
            renderer.join()
        clear_queue(grid_render_queue)


def render_frames(
    input: Union[Queue, mp_Queue],
    output: Union[Queue, mp_Queue],
    ready: Union[Event, mp_Event],
    ImageClass: type,
    style_spec: str,
    repeat: int,
    cached: Union[bool, int],
):
    """Renders animation frames.

    Intended to be executed in a subprocess or thread.
    """
    from ..image import ImageIterator

    image = animator = None  # Silence flake8's F821
    block = True
    while True:
        try:
            data, size, alpha = input.get(block)
        except Empty:
            try:
                output.put(
                    (
                        next(animator),
                        animator._loop_no,
                        image.tell(),
                        size,
                        image.rendered_size,
                    )
                )
            except StopIteration:
                output.put((None,) * 5)
                block = True
            except Exception as e:
                output.put((None,) * 5)
                logging.log_exception(
                    (
                        f"Failed to render frame {image.tell()} of {image._source!r}"
                        + (f" during loop {animator.loop_no}" if repeat != -1 else "")
                    ),
                    logger,
                )
                notify.notify(str(e), level=notify.ERROR)
                block = True
        else:
            if not data:
                break

            if data is ...:
                try:
                    animator.close()
                except AttributeError:  # First time
                    pass
                clear_queue(output)
                ready.set()
                block = True
            elif isinstance(data, tuple):
                new_repeat, frame_no = data
                animator = ImageIterator(
                    image, new_repeat, f"1.1{alpha}{style_spec}", cached
                )
                next(animator)
                animator.seek(frame_no)
                image.set_size(Size.AUTO, maxsize=size)
                block = False
            else:
                # A new image is always created to ensure:
                # 1. the seek position of the image
                #    in MainProcess::MainThread is always correct, since frames should
                #    be rendered ahead.
                # 2. the image size is not changed from another thread in the course of
                #    animation (could occur when an animated image is opened from a
                #    grid wherein its cell is yet to be rendered, since GridRenderer
                #    will continue rendering cells alongside the animation).
                image = ImageClass.from_file(data)
                animator = ImageIterator(
                    image, repeat, f"1.1{alpha}{style_spec}", cached
                )
                image.set_size(Size.AUTO, maxsize=size)
                block = False

    clear_queue(output)


def render_images(
    input: Union[Queue, mp_Queue],
    output: Union[Queue, mp_Queue],
    ImageClass: type,
    style_spec: str,
    *,
    out_extras: bool,
    log_faults: bool,
):
    """Renders images.

    Args:
        out_extras: If True, details other than the render output and it's size are
          also passed out.
    Intended to be executed in a subprocess or thread.
    """
    while True:
        if log_faults:
            image, size, alpha, faulty = input.get()
        else:
            image, size, alpha = input.get()

        if not image:  # Quitting
            break

        image = ImageClass.from_file(image)
        image.set_size(Size.AUTO, maxsize=size)

        # Using `BaseImage` for padding will use more memory since all the
        # spaces will be in the render output string, and theoretically more time
        # with all the checks and string splitting & joining.
        # While `ImageCanvas` is better since it only stores the main image render
        # string (as a list though) then generates and yields the complete lines
        # **as needed**. Trimmed padding lines are never generated at all.
        try:
            output.put(
                (
                    image._source,
                    f"{image:1.1{alpha}{style_spec}}",
                    size,
                    image.rendered_size,
                )
                if out_extras
                else (f"{image:1.1{alpha}{style_spec}}", image.rendered_size)
            )
        except Exception as e:
            output.put(
                (image._source, None, size, image.rendered_size)
                if out_extras
                else (None, image.rendered_size)
            )
            # *faulty* ensures a fault is logged only once per `Image` instance
            if log_faults:
                if not faulty:
                    logging.log_exception(
                        f"Failed to load or render {image._source!r}",
                        logger,
                    )
                notify.notify(str(e), level=notify.ERROR)

    clear_queue(output)


logger = _logging.getLogger(__name__)
anim_render_queue = Queue()
grid_render_queue = Queue()
image_render_queue = Queue()

# Updated from `.tui.init()`
anim_style_specs = {"kitty": "+W", "iterm2": "+Wm1"}
grid_style_specs = {"kitty": "+L", "iterm2": "+L"}
image_style_specs = {"kitty": "+W", "iterm2": "+W"}

# Set from `.tui.init()`
# # Corresponsing to command-line args
ANIM_CACHED: Union[None, bool, int] = None
FRAME_DURATION: Optional[float] = None
REPEAT: Optional[int] = None



"""Issuing user notifications in the TUI and on STDOUT"""

from __future__ import annotations

import logging as _logging
from queue import Queue
from sys import stderr, stdout
from threading import Event, Thread
from time import sleep
from typing import Any, Optional, Tuple, Union

import urwid

from . import cli, logging, tui
from .config import config_options
from .tui import main, widgets
from .utils import COLOR_RESET, CSI

DEBUG = INFO = 0
WARNING = 1
ERROR = 2
CRITICAL = 3


def add_notification(msg: Union[str, Tuple[str, str]]) -> None:
    """Adds a message to the TUI notification bar."""
    if _alarms.full():
        clear_notification(main.loop, None)
    widgets.notifications.contents.insert(
        0, (urwid.Filler(urwid.Text(msg, wrap="ellipsis")), ("given", 1))
    )
    _alarms.put(main.loop.set_alarm_in(5, clear_notification))


def clear_notification(
    loop: Union[urwid.MainLoop, urwid.main_loop.EventLoop], data: Any
) -> None:
    """Removes the oldest message in the TUI notification bar."""
    widgets.notifications.contents.pop()
    loop.remove_alarm(_alarms.get())


def end_loading() -> None:
    """Signals the end of all progressive operations for the current mode."""
    global _n_loading

    if not logging.QUIET:
        _n_loading = -1
        _loading.set()


def is_loading() -> bool:
    """Returns ``True`` if the loading indicator is active or ``False`` if not."""
    return _loading.is_set()


def load() -> None:
    """Displays a loading indicator.

    - elipsis-style for the CLI
    - braille-style for the TUI
    """
    from .tui.main import update_screen
    from .tui.widgets import loading

    global _n_loading

    stream = stdout if stdout.isatty() else stderr
    _loading.wait()

    while _n_loading > -1:
        while _n_loading > 0:
            for stage in (".  ", ".. ", "..."):
                stream.write(stage + "\b" * 3)
                stream.flush()
                if _n_loading <= 0:
                    break
                sleep(0.25)
        stream.write(" " * 3 + "\b" * 3)
        stream.flush()
        if _n_loading > -1:
            _loading.clear()
            _loading.wait()

    _n_loading = 0
    _loading.clear()
    _loading.wait()

    while _n_loading > -1:
        while _n_loading > 0:
            for stage in (
                "\u28bf",
                "\u28fb",
                "\u28fd",
                "\u28fe",
                "\u28f7",
                "\u28ef",
                "\u28df",
                "\u287f",
            ):
                loading.set_text(stage)
                update_screen()
                if _n_loading <= 0:
                    break
                sleep(0.25)
        loading.set_text("")
        update_screen()
        if _n_loading > -1:
            _loading.clear()
            _loading.wait()


def notify(
    msg: str,
    level: int = INFO,
    context: str = "",
    *,
    verbose: bool = False,
    loading: bool = False,
) -> None:
    """Displays a message in the TUI's notification bar or to STDOUT/STDERR."""
    if (
        (cli.args.quiet if logging.QUIET is None else logging.QUIET)
        and level < CRITICAL
        or verbose
        and not (
            cli.args.verbose or cli.args.debug
            if logging.VERBOSE is None
            else logging.VERBOSE
        )
    ):
        return

    if not tui.is_launched:
        print(
            (f"{CSI}34m{context}:{COLOR_RESET} " if context else "")
            + (
                f"{CSI}31m{msg}{COLOR_RESET}"
                if level >= ERROR
                else f"{CSI}33m{msg}{COLOR_RESET}"
                if level == WARNING
                else msg
            ),
            file=stderr if level >= WARNING else stdout,
        )
        if loading:
            start_loading()
    else:
        if config_options.max_notifications:
            add_notification(
                [
                    ("notif context", f"{context}: " if context else ""),
                    # CRITICAL-level notifications should never be displayed in the TUI,
                    # since the program shouldn't recover from the cause.
                    (msg, ("warning", msg), ("error", msg))[level],
                ]
            )


def start_loading() -> None:
    """Signals the start of a progressive operation."""
    global _n_loading

    if not (logging.QUIET or cli.interrupted.is_set() or main.quitting.is_set()):
        _n_loading += 1
        _loading.set()


def stop_loading() -> None:
    """Signals the end of a progressive operation."""
    global _n_loading

    if not logging.QUIET:
        _n_loading -= 1


logger = _logging.getLogger(__name__)

_alarms = Queue(5)  # Max value for "max notifications" is 5

_loading = Event()
_n_loading = 0

# Set from `.logging.init_log()`.
loading_indicator: Optional[Thread] = None
